# -*- coding: utf-8 -*-

"""An internal domain-specific language for probability expressions."""

from __future__ import annotations

import functools
import itertools as itt
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from operator import attrgetter
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

__all__ = [
    "Element",
    "Variable",
    "Intervention",
    "CounterfactualVariable",
    "Distribution",
    "Event",
    "P",
    "Probability",
    "Sum",
    "Product",
    "Fraction",
    "Expression",
    "One",
    "Zero",
    "Q",
    "QFactor",
    "A",
    "AA",
    "B",
    "C",
    "D",
    "M",
    "R",
    "S",
    "T",
    "U",
    "W",
    "X",
    "Y",
    "Z",
    "U1",
    "U2",
    "U3",
    "U4",
    "U5",
    "U6",
    "V1",
    "V2",
    "V3",
    "V4",
    "V5",
    "V6",
    "W0",
    "W1",
    "W2",
    "W3",
    "W4",
    "W5",
    "W6",
    "Y1",
    "Y2",
    "Y3",
    "Y4",
    "Y5",
    "Y6",
    "Z1",
    "Z2",
    "Z3",
    "Z4",
    "Z5",
    "Z6",
    # Helpers
    "ensure_ordering",
    "vmap_adj",
    "vmap_pairs",
    # Transport
    "PopulationProbability",
    "PP",
    "Pi1",
    "Pi2",
    "Pi3",
    "Pi4",
    "Pi5",
    "Pi6",
    "π1",
    "π2",
    "π3",
    "π4",
    "π5",
    "π6",
    "Population",
]

T_co = TypeVar("T_co", covariant=True)


def _to_interventions(variables: Sequence[Variable]) -> Tuple[Intervention, ...]:
    return tuple(
        variable
        if isinstance(variable, Intervention)
        else Intervention(name=variable.name, star=False)
        for variable in variables
    )


class Element(ABC):
    """An element in the y0 internal domain-speific language that can be converted to text, LaTeX, and code."""

    @abstractmethod
    def to_text(self) -> str:
        """Output this DSL object in the internal string format."""

    @abstractmethod
    def to_latex(self) -> str:
        """Output this DSL object in the LaTeX string format."""

    @abstractmethod
    def to_y0(self) -> str:
        """Output this DSL object as y0 python code."""

    def _repr_latex_(self) -> str:  # hack for auto-display of latex in jupyter notebook
        return f"${self.to_latex()}$"

    def __str__(self) -> str:
        return self.to_y0()

    def __repr__(self) -> str:
        return self.to_y0()

    @abstractmethod
    def _iter_variables(self) -> Iterable[Variable]:
        """Iterate over variables."""

    def get_variables(self) -> Set[Variable]:
        """Get the set of variables used in this expression."""
        return set(self._iter_variables())


@dataclass(frozen=True, order=True, repr=False)
class Variable(Element):
    """A variable, typically with a single letter."""

    #: The name of the variable
    name: str
    #: The star status of the variable. None means it's a variable,
    #: False means it's the same as the value for the variable,
    #: and True means it's a different value from the variable.
    star: Optional[bool] = None

    def __post_init__(self):
        if self.name in {"P", "Q", "PP"}:
            raise ValueError(f"trust me, {self.name} is a bad variable name.")

    @classmethod
    def norm(cls, name: Union[str, Variable]) -> Variable:
        """Automatically upgrade a string to a variable."""
        if isinstance(name, str):
            return Variable(name)
        elif isinstance(name, Variable):
            return name
        else:
            raise TypeError(f"({type(name)}) {name} is not valid")

    def get_base(self) -> Variable:
        """Return the base variable, with no other nonsense."""
        return Variable(self.name)

    def to_text(self) -> str:
        """Output this variable in the internal string format."""
        return self.name

    def to_latex(self) -> str:
        """Output this variable in the LaTeX string format.

        :returns: The LaTeX representaton of this variable.

        >>> Variable('X').to_latex()
        'X'
        >>> Variable('X1').to_latex()
        'X_1'
        >>> Variable('X12').to_latex()
        'X_{12}'
        """
        # if it ends with a number, use that as a subscript
        ending_numeric = 0
        for c in reversed(self.name):
            if c.isnumeric():
                ending_numeric += 1
        if ending_numeric == 0:
            return self.name
        elif ending_numeric == 1:
            return f"{self.name[:-1]}_{self.name[-1]}"
        else:
            return f"{self.name[:-ending_numeric]}_{{{self.name[-ending_numeric:]}}}"

    def to_y0(self) -> str:
        """Output this variable instance as y0 internal DSL code."""
        if self.star is None:
            return self.name
        elif self.star:
            return f"+{self.name}"
        else:
            return f"-{self.name}"

    def intervene(self, variables: VariableHint) -> CounterfactualVariable:
        """Intervene on this variable with the given variable(s).

        :param variables: The variable(s) used to extend this variable as it is changed to a
            counterfactual variable
        :returns: A new counterfactual variable over this variable with the given intervention(s).

        .. note:: This function can be accessed with the matmult @ operator.
        """
        interventions = _to_interventions(_upgrade_variables(variables))
        interventions = tuple(sorted(set(interventions), key=lambda i: (i.name, i.star)))
        return CounterfactualVariable(
            name=self.name,
            star=self.star,
            interventions=interventions,
        )

    def __matmul__(self, variables: VariableHint) -> CounterfactualVariable:
        return self.intervene(variables)

    def given(self, parents: Union[VariableHint, Distribution]) -> Distribution:
        """Create a distribution in which this variable is conditioned on the given variable(s).

        The new distribution is a Markov Kernel.

        :param parents: A variable or list of variables to include as conditions in the new conditional distribution
        :returns: A new conditional probability distribution
        :raises TypeError: If a distribution is given as the parents that contains conditionals

        .. note:: This function can be accessed with the or | operator.
        """
        if not isinstance(parents, Distribution):
            return Distribution(
                children=(self,),
                parents=_upgrade_ordering(parents),
            )
        elif parents.is_conditioned():
            raise TypeError("can not be given a distribution that has conditionals")
        else:
            # The parents variable is actually a Distribution instance with no parents,
            #  so its children become the parents for the new Markov Kernel distribution
            return Distribution(
                children=(self,),
                parents=parents.children,  # don't think about this too hard
            )

    def __or__(self, parents: Union[VariableHint, Distribution]) -> Distribution:
        return self.given(parents)

    def joint(self, children: VariableHint) -> Distribution:
        """Create a joint distribution between this variable and the given variable(s).

        :param children: The variable(s) for use with this variable in a joint distribution
        :returns: A new joint distribution over this variable and the given variables.

        .. note:: This function can be accessed with the and & operator.
        """
        return Distribution(
            children=_upgrade_ordering((self, *_upgrade_variables(children))),
        )

    def __and__(self, children: VariableHint) -> Distribution:
        return self.joint(children)

    def _intervention(self, star: bool) -> Variable:
        return Intervention(name=self.name, star=star)

    def invert(self) -> Variable:
        """Create an :class:`Intervention` variable that is different from what was observed (with a star)."""
        return self._intervention(not self.star)

    def __invert__(self) -> Variable:
        return self.invert()

    def __pos__(self) -> Variable:
        return self._intervention(True)

    def __neg__(self) -> Variable:
        return self._intervention(False)

    @classmethod
    def __class_getitem__(cls, item) -> Variable:
        return Variable(item)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get a set containing this variable."""
        yield self


VariableHint = Union[str, Variable, Iterable[Union[str, Variable]]]


@dataclass(frozen=True, order=True, repr=False)
class Intervention(Variable):
    """An intervention variable.

    An intervention variable is usually used as a subscript in a :class:`CounterfactualVariable`.
    """

    def __post_init__(self):
        if self.star is None:
            raise ValueError("Intervention must have a non-None star")

    def to_text(self) -> str:
        """Output this intervention variable in the internal string format."""
        return f"{self.name}*" if self.star else self.name

    def to_latex(self) -> str:
        """Output this intervention variable in the LaTeX string format."""
        latex = super().to_latex()
        return f"{latex}^*" if self.star else latex

    def to_y0(self) -> str:
        """Output this intervention instance as y0 internal DSL code."""
        mark = "+" if self.star else "-"
        return f"{mark}{self.name}"


@dataclass(frozen=True, order=True, repr=False)
class CounterfactualVariable(Variable):
    """A counterfactual variable.

    Counterfactual variables are like normal variables, but can have a list of interventions.
    Each intervention is either the same as what was observed (no star) or different from what
    was observed (star).
    """

    #: The interventions on the variable. Should be non-empty
    interventions: Tuple[Intervention, ...] = field(default_factory=tuple)

    def __post_init__(self):
        if not self.interventions:
            raise ValueError("should give at least one intervention")
        for intervention in self.interventions:
            if not isinstance(intervention, Intervention):
                raise TypeError(
                    f"only Intervention instances are allowed."
                    f" Got: ({intervention.__class__.__name__}) {intervention}",
                )

    def to_text(self) -> str:
        """Output this counterfactual variable in the internal string format."""
        intervention_latex = _list_to_text(self.interventions)
        return f"{self.name}_{{{intervention_latex}}}"

    def to_latex(self) -> str:
        """Output this counterfactual variable in the LaTeX string format.

        :returns: A latex representation of this counterfactual variable

        >>> (Variable('X') @ Variable('Y')).to_latex()
        '{X}_{Y}'
        >>> (Variable('X1') @ Variable('Y')).to_latex()
        '{X_1}_{Y}'
        >>> (Variable('X12') @ Variable('Y')).to_latex()
        '{X_{12}}_{Y}'
        """
        intervention_latex = _list_to_latex(self.interventions)
        prefix = "^*" if self.star else ""
        return f"{{{super().to_latex()}}}{prefix}_{{{intervention_latex}}}"

    def to_y0(self) -> str:
        """Output this counterfactual variable instance as y0 internal DSL code."""
        if self.star is None:
            prefix = ""
        elif self.star:
            prefix = "+"
        else:
            prefix = "-"
        if len(self.interventions) == 1:
            return f"{prefix}{self.name} @ {list(self.interventions)[0].to_y0()}"
        else:
            ins = ", ".join(i.to_y0() for i in self.interventions)
            return f"{prefix}{self.name} @ ({ins})"

    def is_event(self) -> bool:
        """Return if the counterfactual variable has a value."""
        return self.star is not None

    def has_tautology(self) -> bool:
        """Return if the counterfactual variable contain its own value in the subscript.

        :returns: True if we force a variable X to have a value x and the resulting value of X is x.
        :raises ValueError: if the counterfactual value doesn't have a value assigned
        """
        if not self.is_event():
            raise ValueError(
                "Can not determine the consistency of a counterfactual variable with no value assigned."
            )
        return any(self.name == i.name and self.star == i.star for i in self.interventions)

    def is_inconsistent(self) -> bool:
        """Return if the counterfactual variable violates the Axiom of Effectiveness.

        :returns: True if we force a variable X to have a value x and the resulting value of X is not x
        :raises ValueError: if the counterfactual value doesn't have a value assigned
        """
        if not self.is_event():
            raise ValueError(
                "Can not determine the consistency of a counterfactual variable with no value assigned."
            )
        return any(self.name == i.name and self.star != i.star for i in self.interventions)

    def intervene(self, variables: VariableHint) -> CounterfactualVariable:
        """Intervene on this counterfactual variable with the given variable(s).

        :param variables: The variable(s) used to extend this counterfactual variable's
            current interventions. Automatically converts variables to interventions.
        :returns: A new counterfactual variable with both this counterfactual variable's interventions
            and the given intervention(s).

        .. warning::

            Will raise a value error ff the value of a new intervention conflicts
            with the value of intervention already listed in this counterfactual.

        .. note:: This function can be accessed with the matmult @ operator.
        """
        _interventions = _to_interventions(_upgrade_ordering(variables))
        interventions = {*self.interventions, *_interventions}
        self._raise_for_overlapping_interventions(interventions)
        return CounterfactualVariable(
            name=self.name,
            star=self.star,
            interventions=tuple(sorted(interventions, key=attrgetter("name"))),
        )

    @staticmethod
    def _raise_for_overlapping_interventions(interventions: Iterable[Intervention]) -> None:
        """Raise an error if there are two values of the same variable in the list of interventions.

        :param interventions: Interventions to check for overlap
        :raises ValueError: If there are overlapping variables given
        """
        overlaps = {
            (old, new)
            for old, new in itt.product(interventions, repeat=2)
            if old.name == new.name and old.star != new.star
        }
        if overlaps:
            raise ValueError(f"Overlapping interventions in new interventions: {overlaps}")

    def _with_star(self, star: bool) -> CounterfactualVariable:
        return CounterfactualVariable(
            name=self.name,
            star=star,
            interventions=self.interventions,
        )

    def invert(self) -> CounterfactualVariable:
        """Invert the value of the counterfactual variable."""
        return self._with_star(not self.star)

    def __pos__(self) -> CounterfactualVariable:
        return self._with_star(True)

    def __neg__(self) -> CounterfactualVariable:
        return self._with_star(False)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the union of this variable and its interventions."""
        yield from super()._iter_variables()
        for intervention in self.interventions:
            yield from intervention._iter_variables()


@dataclass(frozen=True)
class Distribution(Element):
    """A general distribution over several child variables, conditioned by several parents.

    P(X | Y) means that X is a child and Y is a parent.
    """

    children: Tuple[Variable, ...]
    parents: Tuple[Variable, ...] = field(default_factory=tuple)

    def __post_init__(self):
        if isinstance(self.children, (list, Variable)):
            raise TypeError(f"children of wrong type: {type(self.children)}")
        if isinstance(self.parents, (list, Variable)):
            raise TypeError
        if not self.children:
            raise ValueError("distribution must have at least one child")

    @classmethod
    def safe(
        cls,
        distribution: Union[VariableHint, Distribution],
        *args: Union[str, Variable, Distribution],
    ) -> Distribution:
        """Create a distribution the given variable(s) or distribution.

        :param distribution: If given a :class:`Distribution`, creates a probability expression
            directly over the distribution. If given variable or list of variables, conveniently
            creates a :class:`Distribution` with the variable(s) as children.
        :param args: If the first argument (``distribution``) was given as a single variable, the
            ``args`` variadic argument can be used to specify a list of additional variables.
        :returns: A Distribution object
        :raises ValueError: If invalid combination of arguments are given.
        """
        if isinstance(distribution, (str, Variable, Distribution)):
            extended_args = [distribution, *args]
            dist_pos = [i for i, e in enumerate(extended_args) if isinstance(e, Distribution)]

            # There are no distributions (e.g., no conditionals were given with the | already)
            if 0 == len(dist_pos):
                return Distribution(
                    children=_upgrade_ordering(cast(VariableHint, extended_args)),
                )

            # A single conditional was given. Everything before it should be considered
            # as child variables, and everything after as parent variables.
            elif 1 == len(dist_pos):
                i = dist_pos[0]
                pre = cast(Iterable[Union[str, Variable]], extended_args[:i])
                dist = cast(Distribution, extended_args[i])
                post = cast(Iterable[Union[str, Variable]], extended_args[i + 1 :])
                return Distribution(
                    children=_sorted_variables((*_upgrade_ordering(pre), *dist.children)),
                    parents=_sorted_variables((*dist.parents, *_upgrade_ordering(post))),
                )

            # Multiple conditionals were detected. This isn't allowed.
            else:
                raise ValueError("can not give multiple distribution objects")
        elif args:
            raise ValueError("can not use args/parents when giving an iterable as first argument")
        else:
            return Distribution(
                children=_upgrade_ordering(distribution),
            )

    def _to_x(self, func: Callable[[Iterable[Variable]], str]) -> str:
        children = func(self.children)
        if not self.parents:
            return children
        return f"{children} | {func(self.parents)}"

    def to_text(self) -> str:
        """Output this distribution in the internal string format."""
        return self._to_x(_list_to_text)

    def to_y0(self) -> str:
        """Output this distribution instance as y0 internal DSL code."""
        return self._to_x(_list_to_y0)

    def to_latex(self) -> str:
        """Output this distribution in the LaTeX string format."""
        return self._to_x(_list_to_latex)

    def is_conditioned(self) -> bool:
        """Return if this distribution is conditioned."""
        return 0 < len(self.parents)

    def is_markov_kernel(self) -> bool:
        """Return if this distribution a markov kernel -> one child variable and one or more conditionals."""
        return len(self.children) == 1

    def intervene(self, variables: VariableHint) -> Distribution:
        """Return a new distribution that has the given intervention(s) on all variables."""
        # check that the variables aren't in any of them yet
        variables = _upgrade_ordering(variables)
        return Distribution(
            children=tuple(child.intervene(variables) for child in self.children),
            parents=tuple(parent.intervene(variables) for parent in self.parents),
        )

    def __matmul__(self, variables: VariableHint) -> Distribution:
        return self.intervene(variables)

    def uncondition(self) -> Distribution:
        """Return a new distribution that is not conditioned on the parents."""
        return Distribution(
            children=(*self.children, *self.parents),
        )

    def joint(self, children: VariableHint) -> Distribution:
        """Create a new distribution including the given child variables.

        :param children: The variable(s) with which this distribution's children are extended
        :returns: A new distribution.

        .. note:: This function can be accessed with the and & operator.
        """
        return Distribution(
            children=_upgrade_ordering((*self.children, *_upgrade_variables(children))),
            parents=self.parents,
        )

    def __and__(self, children: VariableHint) -> Distribution:
        return self.joint(children)

    def given(self, parents: Union[VariableHint, Distribution]) -> Distribution:
        """Create a new mixed distribution additionally conditioned on the given parent variables.

        :param parents: The variable(s) with which this distribution's parents are extended
        :returns: A new distribution
        :raises TypeError: If a distribution is given as the parents that contains conditionals

        .. note:: This function can be accessed with the or | operator.
        """
        # TODO handle duplicate variables in the parents.
        if not isinstance(parents, Distribution):
            return Distribution(
                children=self.children,
                parents=_upgrade_ordering((*self.parents, *_upgrade_variables(parents))),
            )
        elif parents.is_conditioned():
            raise TypeError("can not be given a distribution that has conditionals")
        else:
            # The parents variable is actually a Distribution instance with no parents,
            #  so its children get appended as parents for the new mixed distribution
            return Distribution(
                children=self.children,
                parents=(
                    *self.parents,
                    *parents.children,
                ),  # don't think about this too hard
            )

    def __or__(self, parents: Union[VariableHint, Distribution]) -> Distribution:
        return self.given(parents)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in this distribution."""
        for variable in itt.chain(self.children, self.parents):
            yield from variable._iter_variables()


class Expression(Element, ABC):
    """The abstract class representing all expressions."""

    @abstractmethod
    def __mul__(self, other):
        pass

    @abstractmethod
    def _get_key(self) -> tuple:
        """Generate a sort key for a *canonical* expression.

        :returns: A tuple in which the first element is the integer priority for the expression
            and the rest depends on the expression type.
        """
        raise NotImplementedError

    def __lt__(self, other: Expression):
        return self._get_key() < other._get_key()

    def __truediv__(self, expression: Expression) -> Expression:
        """Divide this expression by another and create a fraction."""
        if isinstance(expression, One):
            return self
        elif isinstance(expression, Fraction):
            return Fraction(self * expression.denominator, expression.numerator)
        else:
            return Fraction(self, expression)

    def conditional(self, ranges: VariableHint) -> Expression:
        """Return this expression, conditioned by the given variables.

        :param ranges: A variable or list of variables over which to marginalize this expression
        :returns: A fraction in which the denominator is represents the sum over the given ranges

        >>> from y0.dsl import P, A, B
        >>> assert P(A, B).conditional(A) == P(A, B) / Sum[B](P(A, B))
        >>> assert P(A, B, C).conditional([A, B]) == P(A, B, C) / Sum[C](P(A, B, C))
        """
        ranges = _upgrade_ordering([r.get_base() for r in _upgrade_variables(ranges)])
        ranges_complement = set([c.get_base() for c in self._iter_variables()]) - set(ranges)
        return self.normalize_marginalize(ranges_complement)

    def normalize_marginalize(self, ranges: VariableHint) -> Expression:
        """Return this expression, normalized by this expression marginalized by the given variables."""
        return self / self.marginalize(ranges)

    def marginalize(self, ranges: VariableHint) -> Expression:
        """Return this expression, marginalizing out the given variables.

        :param ranges: A variable or list of variables over which to marginalize this expression
        :returns: The expression but summed over the given variables

        >>> from y0.dsl import P, A, B, C
        >>> assert P(A, B).marginalize(A) == Sum[A](P(A, B))
        >>> assert P(A, B, C).marginalize([A, B]) == Sum[A, B](P(A, B, C))
        """
        return Sum.safe(
            expression=self,
            ranges=_upgrade_ordering([r.get_base() for r in _upgrade_variables(ranges)]),
        )


@dataclass(frozen=True, repr=False)
class Probability(Expression):
    """The probability over a distribution."""

    #: The distribution over which the probability is expressed
    distribution: Distribution

    @classmethod
    def safe(
        cls,
        distribution: DistributionHint,
        *args: Union[str, Variable],
        interventions: Optional[VariableHint] = None,
    ) -> Probability:
        """Create a distribution the given variable(s) or distribution.

        :param distribution: If given a :class:`Distribution`, creates a probability expression
            directly over the distribution. If given variable or list of variables, conveniently
            creates a :class:`Distribution` with the variable(s) as children.
        :param args: If the first argument (``distribution``) was given as a single variable, the
            ``args`` variadic argument can be used to specify a list of additional variables.
        :param interventions: An optional variable or variables to use as interventions.
        :returns: A probability object
        """
        distribution = Distribution.safe(distribution, *args)
        if interventions is not None:
            distribution = distribution.intervene(interventions)
        return Probability(distribution)

    def _get_key(self):
        # TODO incorporate more information from children and parents
        return 0, self.children[0].name

    def to_text(self) -> str:
        """Output this probability in the internal string format."""
        return f"P({self.distribution.to_text()})"

    def _help_level_2_distribution(self):
        # if all parts of distribution have same intervention set, then put it out front
        intervention_sets = {
            x.interventions if isinstance(x, CounterfactualVariable) else tuple()
            for x in itt.chain(self.children, self.parents)
        }
        # check that there's only one intervention set and that it's not an empty one
        if len(intervention_sets) == 1 and (interventions := intervention_sets.pop()):
            unintervened_distribution = Distribution(
                parents=tuple(Variable(name=v.name, star=v.star) for v in self.parents),
                children=tuple(Variable(name=v.name, star=v.star) for v in self.children),
            )
            return interventions, unintervened_distribution
        else:
            return None, None

    def to_y0(self) -> str:
        """Output this probability instance as y0 internal DSL code."""
        interventions, unintervened_distribution = self._help_level_2_distribution()
        if not interventions:
            return f"P({self.distribution.to_y0()})"

        # only keep the + if necessary, otherwise show regular
        intervention_str = ",".join(
            f"+{intervention.name}" if intervention.star else intervention.name
            for intervention in interventions
        )
        return f"P[{intervention_str}]({unintervened_distribution.to_y0()})"

    def to_latex(self) -> str:
        """Output this probability in the LaTeX string format."""
        interventions, unintervened_distribution = self._help_level_2_distribution()
        if not interventions:
            return f"P({self.distribution.to_latex()})"

        intervention_str = ",".join(intervention.to_latex() for intervention in interventions)
        return f"P_{{{intervention_str}}}({unintervened_distribution.to_latex()})"

    @property
    def parents(self) -> Tuple[Variable, ...]:
        """Get the distribution's parents."""
        return self.distribution.parents

    @property
    def children(self) -> Tuple[Variable, ...]:
        """Get the distribution's children."""
        return self.distribution.children

    def is_conditioned(self) -> bool:
        """Return if this distribution is conditioned."""
        return self.distribution.is_conditioned()

    def is_markov_kernel(self) -> bool:
        """Return if this distribution a markov kernel -> one child variable and one or more conditionals."""
        return self.distribution.is_markov_kernel()

    def __mul__(self, other: Expression) -> Expression:
        if isinstance(other, Zero):
            return other
        elif isinstance(other, One):
            return self
        elif isinstance(other, Product):
            return Product.safe((self, *other.expressions))
        elif isinstance(other, Fraction):
            return Fraction(self * other.numerator, other.denominator)
        else:
            return Product.safe((self, other))

    def _new(self, distribution: Distribution):
        # This is implemented this way to make overriding easier
        return Probability(distribution)

    def intervene(self, variables: VariableHint) -> Probability:
        """Return a new probability where the underlying distribution has been intervened by the given variables."""
        return self._new(self.distribution.intervene(variables))

    def __matmul__(self, variables: VariableHint) -> Probability:
        return self.intervene(variables)

    def uncondition(self) -> Probability:
        """Return a new probability where the underlying distribution is no longer conditioned by the parents.

        :returns: A new probability over a distribution over the children and parents of the previous distribution

        >>> from y0.dsl import P, A, B
        >>> P(A | B).uncondition() == P(A, B)
        """
        return self._new(self.distribution.uncondition())

    def conditional(self, ranges: VariableHint) -> Expression:
        """Return this expression, conditioned by the given variables.

        :param ranges: A variable or list of variables over which to marginalize this expression
        :returns: A fraction in which the denominator is represents the sum over the given ranges

        >>> from y0.dsl import P, A, B
        >>> assert P(A, B).conditional(A) == P(A, B) / Sum[B](P(A, B))
        >>> assert P(A, B, C).conditional([A, B]) == P(A, B, C) / Sum[C](P(A, B, C))
        """
        ranges = _upgrade_ordering([r.get_base() for r in _upgrade_variables(ranges)])
        ranges_complement = set(
            [c.get_base() for c in self._iter_variables() if not isinstance(c, Intervention)]
        ) - set(ranges)
        return self.normalize_marginalize(ranges_complement)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in the distribution in this probability."""
        yield from self.distribution._iter_variables()


DistributionHint = Union[VariableHint, Distribution]


class ProbabilityBuilderType:
    """A base class for building probability distributions."""

    def __call__(
        self,
        distribution: DistributionHint,
        *args: Union[str, Variable],
        interventions: Optional[VariableHint] = None,
    ) -> Probability:
        return Probability.safe(distribution, *args, interventions=interventions)

    def __getitem__(self, interventions: VariableHint):
        """Generate a probability builder closure.

        :param interventions: A variable or variables to intervene on using the do-calculus level 2
            rules, meaning they are all applied to all parent and children variables in the resulting
            expression
        :returns: A function with the same semantics as :meth:`__call__` such that you can build
            a probability expression.

        >>> from y0.dsl import P, W, X, Y, Z
        >>> assert P[X](Y) == P(Y @ X)
        >>> assert P[X](Y, Z) == P(Y @ X & Z @ X)
        >>> assert P[X](Y | Z) == P(Y @ X | Z @ X)
        >>> assert P[X](Y @ Z) == P(Y @ Z @ X)
        >>> assert P[X](Y @ Z | W) == P(Y @ Z @ X | W @ X)
        """
        return functools.partial(self, interventions=interventions)


P = ProbabilityBuilderType()
"""``P`` is a magical object of mystery and wonder that can be used to create :class:`Probability` instances.

It itself is a singleton instance of :class:`ProbabilityBuilderType` and can be used wither via the
:meth:`ProbabilityBuilderType.__call__`, as if it were a function like ``P(Y)`` or it can be used as
a combination with the :meth:`ProbabilityBuilderType.__getitem__` and a call, like ``P[X](Y)`` to
denote interventions using the do-Calculus $L_2$ notation. Here are some examples:

A univariate distribution can be created either with a string or a :class:`Variable`:

>>> from y0.dsl import P, A
>>> P('A') == P(A)

**Multivariate Distributions**

A joint distribution can be created with several strings or :class:`Variable` instances
with variadic arguments:

>>> from y0.dsl import P, A, B
>>> P(A, B) == P('A', 'B')

A joint distribution can also be created with a single argument that is either an iterable
of either strings or :class:`Variable` instances

>>> from y0.dsl import P, A, B
>>> P((A, B)) == P([A, B]) == P(('A', 'B')) == P(['A', 'B'])

This even extends to fancy generators, for which you can omit the parentheses:

Creation with a fancy generator of variables:

>>> from y0.dsl import P, A, B
>>> P(Variable(name) for name in 'AB') == P(name for name in 'AB') == P(A, B)

**Conditional Distributions**

Creation with a conditional distribution:

>>> from y0.dsl import P, A, B
>>> P(A | B)

Creation with a mixed joint/conditional distribution:

>>> from y0.dsl import P, A, B, C
>>> P(A & B | C)

**Specifying an Intervention with L2 do-Calculus Notation**

Intervene on a single variable:

>>> from y0.dsl import P, X, Y
>>> P[X](Y) == P(Y @ X)

Intervene on multiple children:

>>> from y0.dsl import P, X, Y, Z
>>> P[X](Y, Z) == P(Y @ X & Z @ X)

Intervene on multiple parents:

>>> from y0.dsl import P, W, X, Y, Z
>>> P[X](Y | (W, Z)) == P(Y @ X | (W @ X, Z @ X)):

Intervene on both children and parents:

>>> from y0.dsl import P, X, Y, Z
>>> P[X](Y | Z) == P(Y @ X | Z @ X)

Intervene on X on top of previous interventions:

>>> from y0.dsl import P, X, Y, Z
>>> P[X](Y @ Z) == P(Y @ X @ Z)

Allow mixing with L3, where each variable can have different interventions:

>>> from y0.dsl import P, W, X, Y, Z
>>> P[X](Y @ Z | W) == P(Y @ X @ Z | W @ X)

**Specifying Multiple Interventions with L2 do-Calculus Notation**

Multiple interventions on a single variable:

>>> from y0.dsl import P, X1, X2, Y
>>> P[X1, X2](Y) == P(Y @ X)

Multiple interventions  on multiple children:

>>> from y0.dsl import P, X1, X2, Y, Z
>>> P[X1, X2](Y, Z) == P(Y @ X1 @ X2 & Z @ X1 @ X2)

... and so on
"""


@dataclass(frozen=True, repr=False)
class Product(Expression):
    """Represent the product of several probability expressions."""

    expressions: Tuple[Expression, ...]

    def __post_init__(self):
        if len(self.expressions) < 2:
            raise ValueError("Product() must two or more expressions")

    @classmethod
    def safe(cls, expressions: Union[Expression, Iterable[Expression]]) -> Expression:
        """Construct a product from any iterable of expressions.

        :param expressions: An expression or iterable of expressions which should be multiplied
        :returns: A :class:`Product` object

        Standard usage, same as the normal ``__init__``:

        >>> from y0.dsl import Product, X, Y, A, P
        >>> Product.safe((P(X, Y), ))

        Use a list or other iterable:

        >>> Product.safe([P(X), P(Y | X)])

        Use an inline generator:

        >>> Product.safe(P(v) for v in [X, Y])

        Use a single expression:

        >>> Product.safe(P(X, Y))
        """
        if isinstance(expressions, Expression):
            return expressions
        # Remove multiplications of one
        expressions = tuple(expression for expression in expressions if expression != One())
        # If any multiplications are by zero, then return zero
        if any(expression == Zero() for expression in expressions):
            return Zero()
        if not expressions:
            return One()
        if len(expressions) == 1:
            return expressions[0]
        return cls(expressions=tuple(sorted(expressions)))

    def _get_key(self):
        inner_keys = (sexpr._get_key() for sexpr in self.expressions)
        return 2, *inner_keys

    def to_text(self):
        """Output this product in the internal string format."""
        return " ".join(expression.to_text() for expression in self.expressions)

    def to_y0(self) -> str:
        """Output this product instance as y0 internal DSL code."""
        return " * ".join(expr.to_y0() for expr in self.expressions)

    def to_latex(self):
        """Output this product in the LaTeX string format."""
        return " ".join(expression.to_latex() for expression in self.expressions)

    def __mul__(self, other: Expression):
        if isinstance(other, Zero):
            return other
        if isinstance(other, Product):
            return Product.safe((*self.expressions, *other.expressions))
        elif isinstance(other, Fraction):
            return Fraction(self * other.numerator, other.denominator)
        else:
            return Product.safe((*self.expressions, other))

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the union of the variables used in each expresison in this product."""
        for expression in self.expressions:
            yield from expression._iter_variables()


def _list_to_text(elements: Iterable[Element]) -> str:
    return ", ".join(element.to_text() for element in elements)


def _list_to_latex(elements: Iterable[Element]) -> str:
    return ", ".join(element.to_latex() for element in elements)


def _list_to_y0(elements: Iterable[Element]) -> str:
    return ", ".join(element.to_y0() for element in elements)


@dataclass(frozen=True, repr=False)
class Sum(Expression):
    """Represent the sum over an expression over an optional set of variables."""

    #: The expression over which the sum is done
    expression: Expression
    #: The variables over which the sum is done. Defaults to an empty list, meaning no variables.
    ranges: frozenset[Variable]

    def __post_init__(self):
        if not isinstance(self.ranges, frozenset):
            raise TypeError
        if not self.ranges:
            raise ValueError("Sum must have ranges")
        for r in self.ranges:
            if isinstance(r, (CounterfactualVariable, Intervention)):
                raise TypeError("Ranges must not be counterfactuals nor interventions")

    @classmethod
    def safe(
        cls,
        expression: Expression,
        ranges: Union[str, Variable, Iterable[Union[str, Variable]]],
        *,
        simplify: bool = False,
    ) -> Expression:
        """Construct a sum from an expression and a permissive set of things in the ranges.

        :param expression: The expression over which the sum is done
        :param ranges: The variable or list of variables over which the sum is done
        :param simplify: Should the sum be simplified using :func:`Sum.simplify`?
        :returns: A :class:`Sum` object

        Standard usage, same as the normal ``__init__``:

        >>> from y0.dsl import Sum, X, Y, A, P
        >>> Sum.safe(P(X, Y), (X,))

        Use a list or other iterable:

        >>> Sum.safe(P(X, Y), [X])

        Use a single variable:

        >>> Sum.safe(P(X, Y), X)
        """
        if isinstance(ranges, str):
            ranges = (Variable(ranges),)
        elif isinstance(ranges, Variable):
            ranges = (ranges,)
        else:
            ranges = _upgrade_ordering(ranges)
        if not ranges:
            return expression
        if isinstance(expression, Zero):
            return expression
        rv = cls(
            expression=expression,
            ranges=frozenset(ranges),
        )
        if simplify:
            return rv.simplify()
        return rv

    def simplify(self) -> Expression:
        """Simplify this sum."""
        expression = self.expression
        ranges = set(self.ranges)

        # Special case when ranges cover
        if isinstance(expression, Probability) and not expression.parents:  # i.e., no conditions
            children = {
                child.get_base(): child
                for child in expression.children
                # FIXME what happens if same name appears with multiple different counterfactual variables?
                #  this should actually evaluate to zero since that's impossible
            }
            if ranges == set(children):
                return One()
            elif ranges > set(children):
                keep = ranges - set(children)
                return Sum.safe(
                    expression=One(),
                    ranges=frozenset(v for k, v in children.items() if k in keep),
                )
            elif ranges < set(children):
                keep = set(children) - ranges
                return expression._new(
                    Distribution.safe(v for k, v in children.items() if k in keep)
                )
            else:  # partial or no overlap
                intersection = ranges.intersection(children)
                keep = set(children) - intersection
                prob = expression._new(
                    Distribution.safe(v for k, v in children.items() if k in keep)
                )
                return Sum.safe(
                    expression=prob,
                    ranges=ranges - intersection,
                )
        return self

    def _get_key(self):
        return 1, *self.expression._get_key()

    def _get_sorted_ranges(self) -> Sequence[Variable]:
        return sorted(self.ranges, key=attrgetter("name"))

    def to_text(self) -> str:
        """Output this sum in the internal string format."""
        ranges = _list_to_text(self._get_sorted_ranges())
        return f"[ sum_{{{ranges}}} {self.expression.to_text()} ]"

    def to_latex(self) -> str:
        """Output this sum in the LaTeX string format."""
        ranges = _list_to_latex(self._get_sorted_ranges())
        return rf"\sum\limits_{{{ranges}}} {self.expression.to_latex()}"

    def to_y0(self):
        """Output this sum instance as y0 internal DSL code."""
        if isinstance(self.expression, Fraction):
            s = self.expression.to_y0(parens=False)
        else:
            s = self.expression.to_y0()
        if not self.ranges:
            return f"Sum({s})"
        ranges = _list_to_y0(self._get_sorted_ranges())
        return f"Sum[{ranges}]({s})"

    def __mul__(self, expression: Expression):
        if isinstance(expression, Zero):
            return expression
        elif isinstance(expression, Product):
            return Product.safe((self, *expression.expressions))
        else:
            return Product.safe((self, expression))

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the union of the variables used in the range of this sum and variables in its summand."""
        yield from self.expression._iter_variables()
        for variable in self.ranges:
            yield from variable._iter_variables()

    @classmethod
    def __class_getitem__(cls, ranges: VariableHint) -> Callable[[Expression], Expression]:
        """Create a partial sum object over the given ranges.

        :param ranges: The variables over which the partial sum will be done
        :returns: A partial :class:`Sum` that can be called solely on an expression

        Example single variable sum:

        >>> from y0.dsl import Sum, P, A, B
        >>> Sum[B](P(A | B) * P(B))

        Example multiple variable sum:

        >>> from y0.dsl import Sum, P, A, B, C
        >>> Sum[B, C](P(A | B) * P(B))
        """
        return functools.partial(Sum.safe, ranges=_upgrade_ordering(ranges))


@dataclass(frozen=True, repr=False)
class Fraction(Expression):
    """Represents a fraction of two expressions."""

    #: The expression in the numerator of the fraction
    numerator: Expression
    #: The expression in the denominator of the fraction
    denominator: Expression

    def __post_init__(self):
        if isinstance(self.denominator, Zero):
            raise ZeroDivisionError

    def _get_key(self):
        return (
            3,
            self.numerator._get_key(),
            self.denominator._get_key(),
        )

    def to_text(self) -> str:
        """Output this fraction in the internal string format."""
        return f"frac_{{{self.numerator.to_text()}}}{{{self.denominator.to_text()}}}"

    def to_latex(self) -> str:
        """Output this fraction in the LaTeX string format."""
        return rf"\frac{{{self.numerator.to_latex()}}}{{{self.denominator.to_latex()}}}"

    def to_y0(self, parens: bool = True) -> str:
        """Output this fraction as y0 internal DSL code."""
        s = f"({self.numerator.to_y0()} / {self.denominator.to_y0()})"
        return f"({s})" if parens else s

    def __mul__(self, expression: Expression) -> Expression:
        if isinstance(expression, Zero):
            return expression
        elif isinstance(expression, Fraction):
            return Fraction(
                self.numerator * expression.numerator,
                self.denominator * expression.denominator,
            )
        else:
            return Fraction(self.numerator * expression, self.denominator)

    def __truediv__(self, expression: Expression) -> Fraction:
        if isinstance(expression, One):
            return self
        elif isinstance(expression, Fraction):
            return Fraction(
                self.numerator * expression.denominator,
                self.denominator * expression.numerator,
            )
        else:
            return Fraction(self.numerator, self.denominator * expression)

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in the numerator and denominator of this fraction."""
        yield from self.numerator._iter_variables()
        yield from self.denominator._iter_variables()

    def flip(self) -> Fraction:
        """Exchange the numerator and denominator."""
        return Fraction(self.denominator, self.numerator)

    def simplify(self) -> Expression:
        """Simplify this fraction."""
        if isinstance(self.denominator, One):
            return self.numerator
        if isinstance(self.numerator, Zero):
            return self.numerator
        if isinstance(self.numerator, One):
            if isinstance(self.denominator, Fraction):
                return self.denominator.flip().simplify()
            else:
                return self
        if self.numerator == self.denominator:
            return One()
        if isinstance(self.numerator, Product) and isinstance(self.denominator, Product):
            return self._simplify_parts(self.numerator.expressions, self.denominator.expressions)
        elif isinstance(self.numerator, Product):
            return self._simplify_parts(self.numerator.expressions, [self.denominator])
        elif isinstance(self.denominator, Product):
            return self._simplify_parts([self.numerator], self.denominator.expressions)
        return self

    @classmethod
    def _simplify_parts(
        cls, numerator: Sequence[Expression], denominator: Sequence[Expression]
    ) -> Expression:
        """Calculate the minimum fraction.

        :param numerator: A sequence of expressions that are multiplied in the product in the numerator
        :param denominator: A sequence of expressions that are multiplied in the product in the denominator
        :returns: A simplified fraction.
        """
        new_numerator, new_denominator = cls._simplify_parts_helper(numerator, denominator)
        if new_numerator and new_denominator:
            return Fraction(
                Product.safe(new_numerator),
                Product.safe(new_denominator),
            )
        elif new_numerator:
            return Product.safe(new_numerator)
        elif new_denominator:
            return One() / Product.safe(new_denominator)
        else:
            return One()

    @staticmethod
    def _simplify_parts_helper(
        numerator: Sequence[Expression],
        denominator: Sequence[Expression],
    ) -> Tuple[Tuple[Expression, ...], Tuple[Expression, ...]]:
        numerator_cancelled = set()
        denominator_cancelled = set()
        for i, n_expr in enumerate(numerator):
            for j, d_expr in enumerate(denominator):
                if j in denominator_cancelled:
                    continue
                if n_expr == d_expr:
                    numerator_cancelled.add(i)
                    denominator_cancelled.add(j)
                    break
        return (
            tuple(expr for i, expr in enumerate(numerator) if i not in numerator_cancelled),
            tuple(expr for i, expr in enumerate(denominator) if i not in denominator_cancelled),
        )


class One(Expression):
    """The multiplicative identity (1)."""

    def to_text(self) -> str:
        """Output this identity variable in the internal string format."""
        return "1"

    def to_latex(self) -> str:
        """Output this identity instance in the LaTeX string format."""
        return "1"

    def to_y0(self) -> str:
        """Output this identity instance as y0 internal DSL code."""
        return "One()"

    def _get_key(self):
        return 4, self.to_text()

    def __rmul__(self, expression: Expression) -> Expression:
        return expression

    def __mul__(self, expression: Expression) -> Expression:
        return expression

    def __eq__(self, other):
        return isinstance(other, One)  # all ones are equal

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in this expression."""
        return iter([])


class Zero(Expression):
    """The additive identity (0)."""

    def to_text(self) -> str:
        """Output this identity variable in the internal string format."""
        return "0"

    def to_latex(self) -> str:
        """Output this identity instance in the LaTeX string format."""
        return "0"

    def to_y0(self) -> str:
        """Output this identity instance as y0 internal DSL code."""
        return "Zero()"

    def _get_key(self):
        return 4, self.to_text()

    def __rmul__(self, expression: Expression) -> Expression:
        return self

    def __mul__(self, expression: Expression) -> Expression:
        return self

    def __truediv__(self, other: Expression) -> Expression:
        if isinstance(other, Zero):
            raise ZeroDivisionError
        return self

    def __eq__(self, other):
        return isinstance(other, Zero)  # all zeros are equal

    def _iter_variables(self) -> Iterable[Variable]:
        """Get the set of variables used in this expression."""
        return iter([])


class QBuilder(Protocol[T_co]):
    """A protocol for annotating the special class getitem functionality of the :class:`QFactor` class."""

    def __call__(self, arg: VariableHint, *args: Union[str, Variable]) -> T_co:
        ...


@dataclass(frozen=True, repr=False)
class QFactor(Expression):
    """A function from the variables in the domain to a probability function over variables in the codomain."""

    domain: frozenset[Variable]
    codomain: frozenset[Variable]

    @classmethod
    def safe(
        cls,
        domain: VariableHint,
        *args: Union[str, Variable],
        codomain: VariableHint,
    ) -> QFactor:
        """Create a Q factor with various input types."""
        return cls(
            domain=cls._prepare_domain(domain, *args),
            codomain=frozenset(_upgrade_variables(codomain)),
        )

    @staticmethod
    def _prepare_domain(
        arg: VariableHint,
        *args: Union[str, Variable],
    ) -> frozenset[Variable]:
        """Prepare a list of variables from a potentially unruly set of args and variadic args."""
        if isinstance(arg, (str, Variable)):
            return frozenset((Variable.norm(arg), *_upgrade_ordering(args)))
        if args:
            raise ValueError("can not use variadic arguments with combination of first arg")
        return frozenset(_sorted_variables(_upgrade_ordering(arg)))

    @classmethod
    def __class_getitem__(cls, codomain: Union[Variable, Iterable[Variable]]) -> QBuilder[QFactor]:
        """Create a partial Q Factor object over the given codomain.

        :param codomain: The variables over which the partial Q Factor will be done
        :returns: A partial :class:`QFactor` that can be called solely on an expression

        Example single variable codomain Q expression:

        >>> from y0.dsl import Sum, Q, A, B, C
        >>> Q[C](A, B)

        Example multiple variable codomain Q expression:

        >>> from y0.dsl import Sum, Q, A, B, C, D
        >>> Q[C, D](A, B)
        """
        return functools.partial(cls.safe, codomain=codomain)

    def _get_key(self) -> tuple:
        return -5, min(v.name for v in self.domain), min(v.name for v in self.codomain)

    def _sorted_codomain(self):
        return sorted(self.codomain, key=attrgetter("name"))

    def _sorted_domain(self):
        return sorted(self.domain, key=attrgetter("name"))

    def to_text(self) -> str:
        """Output this Q factor in the internal string format."""
        codomain = _list_to_text(self._sorted_codomain())
        domain = _list_to_text(self._sorted_domain())
        return f"Q[{codomain}]({domain})"

    def to_latex(self) -> str:
        """Output this Q factor in the LaTeX string format."""
        codomain = _list_to_latex(self._sorted_codomain())
        domain = _list_to_latex(self._sorted_domain())
        return rf"Q_{{{codomain}}}({{{domain}}})"

    def to_y0(self) -> str:
        """Output this Q factor instance as y0 internal DSL code."""
        codomain = _list_to_y0(self._sorted_codomain())
        domain = _list_to_y0(self._sorted_domain())
        return f"Q[{codomain}]({domain})"

    def __mul__(self, other: Expression):
        if isinstance(other, Product):
            return Product.safe((self, *other.expressions))
        elif isinstance(other, Fraction):
            return Fraction(self * other.numerator, other.denominator)
        else:
            return Product.safe((self, other))

    def _iter_variables(self) -> Iterable[Variable]:
        yield from self.codomain
        yield from self.domain


Q = QFactor

AA = Variable("AA")
A, B, C, D, E, F, G, M, R, S, T, U, W, X, Y, Z = map(Variable, "ABCDEFGMRSTUWXYZ")  # type: ignore
U1, U2, U3, U4, U5, U6 = [Variable(f"U{i}") for i in range(1, 7)]
V1, V2, V3, V4, V5, V6 = [Variable(f"V{i}") for i in range(1, 7)]
W0, W1, W2, W3, W4, W5, W6 = [Variable(f"W{i}") for i in range(7)]
X1, X2, X3, X4, X5, X6 = [Variable(f"X{i}") for i in range(1, 7)]
Y1, Y2, Y3, Y4, Y5, Y6 = [Variable(f"Y{i}") for i in range(1, 7)]
Z1, Z2, Z3, Z4, Z5, Z6 = [Variable(f"Z{i}") for i in range(1, 7)]
π1, π2, π3, π4, π5, π6 = Pi1, Pi2, Pi3, Pi4, Pi5, Pi6 = [Variable(f"π{i}") for i in range(1, 7)]


def _sorted_variables(variables: Iterable[Variable]) -> Tuple[Variable, ...]:
    return tuple(sorted(variables, key=attrgetter("name")))


def _upgrade_variables(variables: VariableHint) -> Tuple[Variable, ...]:
    if isinstance(variables, str):
        return (Variable(variables),)
    elif isinstance(variables, Variable):
        return (variables,)
    else:
        return tuple(Variable.norm(variable) for variable in variables)


def _upgrade_ordering(variables: VariableHint) -> Tuple[Variable, ...]:
    return _sorted_variables(set(_upgrade_variables(variables)))


OrderingHint = Optional[Iterable[Union[str, Variable]]]


def ensure_ordering(
    expression: Expression,
    *,
    ordering: OrderingHint = None,
) -> Sequence[Variable]:
    """Get a canonical ordering of the variables in the expression, or pass one through.

    The canonical ordering of the variables in a given expression is based on the alphabetical
    sort order of the variables based on their names.

    :param expression: The expression to get a canonical ordering from.
    :param ordering: A given ordering to pass through if not none, otherwise calculate it.
    :returns: The ordering
    """
    if ordering is not None:
        return _upgrade_ordering(ordering)
    # use alphabetical ordering
    return _sorted_variables(expression.get_variables())


def _get_treatment_variables(variables: set[Variable]) -> set[Variable]:
    return {variable for variable in variables if isinstance(variable, Intervention)}


def _get_outcome_variables(variables: set[Variable]) -> set[Variable]:
    return {variable for variable in variables if not isinstance(variable, Intervention)}


def get_outcomes_and_treatments(*, query: Expression) -> tuple[set[Variable], set[Variable]]:
    """Get outcomes and treatments sets from the query expression."""
    variables = query.get_variables()
    return (
        _get_outcome_variables(variables),
        _get_treatment_variables(variables),
    )


def outcomes_and_treatments_to_query(
    *, outcomes: set[Variable], treatments: Optional[set[Variable]] = None
) -> Expression:
    """Create a query expression from a set of outcome and treatment variables."""
    if not treatments:
        return P(outcomes)
    return P(Variable.norm(y) @ _upgrade_ordering(treatments) for y in outcomes)


def vmap_pairs(edges: Iterable[Tuple[str, str]]) -> List[Tuple[Variable, Variable]]:
    """Map pair of strings to pairs of variables."""
    return [(Variable(source), Variable(target)) for source, target in edges]


def vmap_adj(adjacency_dict):
    """Map an adjacency dictionary of strings to variables."""
    return {
        Variable(source): [Variable(target) for target in targets]
        for source, targets in adjacency_dict.items()
    }


#: A conjunction of factual and counterfactual events
Event = Dict[Variable, Intervention]

Population = Variable


@dataclass(frozen=True, repr=False)
class PopulationProbability(Probability):
    """A probability that is annotated with a population.

    >>> from y0.dsl import PP, Pi1, Y, X
    >>> # Make a population-annotated probability of Y
    >>> PP[Pi1](Y)
    >>> # Make a conditioned population of Y @ X
    >>> PP[Pi1][X](Y)

    Related publications:
    - `Surrogate Outcomes and Transportability <https://arxiv.org/abs/1806.07172>`_ (Tikka and Karvanen, 2018)
    """

    population: Population

    def _new(self, distribution) -> PopulationProbability:
        return PopulationProbability(population=self.population, distribution=distribution)

    def _get_key(self):
        return -1, self.population, self.children[0].name

    def to_y0(self) -> str:
        """Output this probability instance as y0 internal DSL code."""
        interventions, unintervened_distribution = self._help_level_2_distribution()
        if not interventions:
            return f"P({self.distribution.to_y0()})"

        # only keep the + if necessary, otherwise show regular
        intervention_str = ",".join(
            f"+{intervention.name}" if intervention.star else intervention.name
            for intervention in interventions
        )
        return f"PP[{self.population.to_y0()}][{intervention_str}]({unintervened_distribution.to_y0()})"

    def to_latex(self) -> str:
        """Output this probability in the LaTeX string format."""
        interventions, unintervened_distribution = self._help_level_2_distribution()
        if self.population == TARGET_DOMAIN:
            pop_latex = r"\pi^\ast"
        else:
            pop_latex = self.population.to_latex()

        if not interventions:
            return f"P^{{{pop_latex}}}({self.distribution.to_latex()})"

        intervention_str = ",".join(intervention.to_latex() for intervention in interventions)
        return f"P_{{{intervention_str}}}^{{{pop_latex}}}({unintervened_distribution.to_latex()})"


class PopulationProbabilityBuilderType(ProbabilityBuilderType):
    """A magical type for building population probabilities."""

    def __init__(self, population: Population):
        """Initialize the builder with a given population."""
        self.population = population

    @classmethod
    def __class_getitem__(cls, population: Population) -> "PopulationProbabilityBuilderType":
        """Get a population probability builder class initialized with the given population."""
        return cls(population)

    def __call__(self, *args, **kwargs) -> PopulationProbability:  # noqa:D102
        probability = super().__call__(*args, **kwargs)
        return PopulationProbability(
            population=self.population, distribution=probability.distribution
        )


PP = PopulationProbabilityBuilderType


TARGET_DOMAIN = Population("pi*")



# -*- coding: utf-8 -*-

"""Tests for the ID* algorithm."""

from tests.test_algorithm import cases
from y0.algorithm.identify._extras import (
    get_district_interventions,
    intervene_on_district,
)
from y0.algorithm.identify.cg import make_counterfactual_graph
from y0.algorithm.identify.id_star import (
    ConflictUnidentifiable,
    get_cf_interventions,
    get_conflicts,
    get_events_of_district,
    get_events_of_each_district,
    get_evidence,
    get_free_variables,
    id_star,
    id_star_line_6,
    id_star_line_9,
    is_redundant_counterfactual,
    remove_event_tautologies,
    violates_axiom_of_effectiveness,
)
from y0.algorithm.identify.idc_star import (
    cf_rule_2_of_do_calculus_applies,
    get_new_outcomes_and_conditions,
    get_remaining_and_missing_events,
    idc_star,
)
from y0.dsl import D, One, P, Sum, W, X, Y, Z, Zero
from y0.examples import (
    figure_9a,
    figure_9c,
    figure_9d,
    tikka_figure_2,
    tikka_figure_5,
    tikka_figure_6a,
    tikka_figure_6b,
    tikka_unidentifiable_cfgraph,
)
from y0.graph import NxMixedGraph

d, w, x, y, z = -D, -W, -X, -Y, -Z


class TestIDCStar(cases.GraphTestCase):
    """Tests for the  ``IDC*`` algorithm."""

    def test_rule_2_of_do_calculus_applies(self):
        """Test that rule 2 of do calculus applies."""
        input_graph1 = NxMixedGraph.from_edges(directed=[(X, Y), (Z, X), (Z, Y)])
        outcomes = {Y}
        condition = X
        self.assertFalse(cf_rule_2_of_do_calculus_applies(input_graph1, outcomes, condition))
        input_graph2 = NxMixedGraph.from_edges(directed=[(X, Y)])
        self.assertTrue(cf_rule_2_of_do_calculus_applies(input_graph2, outcomes, condition))
        self.assertTrue(cf_rule_2_of_do_calculus_applies(input_graph1, outcomes, Z))
        self.assertFalse(
            cf_rule_2_of_do_calculus_applies(tikka_figure_2.graph, outcomes, condition)
        )
        self.assertTrue(cf_rule_2_of_do_calculus_applies(tikka_figure_5.graph, {Y @ -x}, D))
        self.assertFalse(cf_rule_2_of_do_calculus_applies(tikka_figure_6a.graph, {Y}, Z))
        self.assertFalse(cf_rule_2_of_do_calculus_applies(tikka_figure_6a.graph, {Y}, X))
        self.assertFalse(cf_rule_2_of_do_calculus_applies(tikka_figure_6a.graph, {Y @ -x}, X))
        self.assertFalse(cf_rule_2_of_do_calculus_applies(tikka_figure_6b.graph, {Y}, Z))
        self.assertFalse(cf_rule_2_of_do_calculus_applies(tikka_figure_6b.graph, {Y}, X))
        self.assertTrue(cf_rule_2_of_do_calculus_applies(tikka_figure_6a.graph, {Y @ -x}, Z @ -x))


class TestIDStar(cases.GraphTestCase):
    """Tests for the ``ID*`` algorithm."""

    def test_id_star_line_1(self):
        """Check that one is returned when running ID* on an empty event."""
        self.assertEqual(One(), id_star(figure_9a.graph, {}))

    def test_id_star_line_2(self):
        """Check to see if the counterfactual event violates the Axiom of Effectiveness."""
        # Examples all from figure_9a.graph
        self.assertTrue(violates_axiom_of_effectiveness({X @ -x: +x}))
        self.assertTrue(violates_axiom_of_effectiveness({X @ ~x: x}))
        self.assertTrue(violates_axiom_of_effectiveness({X @ (y, z, x): ~x}))
        self.assertTrue(violates_axiom_of_effectiveness({X @ (x, z): x, Y @ (x, -y): +y}))
        self.assertTrue(
            violates_axiom_of_effectiveness(
                {Y @ -x: -y, X @ +x: -x, Y @ x: y, Y @ ~x: y, Y @ ~x: y}
            )
        )
        self.assertFalse(violates_axiom_of_effectiveness({X @ x: x}))
        self.assertFalse(violates_axiom_of_effectiveness({X @ ~x: ~x}))
        self.assertFalse(violates_axiom_of_effectiveness({X @ x: x, Y @ x: y}))
        self.assertFalse(violates_axiom_of_effectiveness({Z @ -x: -z, X: +x}))
        self.assertEqual(Zero(), id_star(figure_9a.graph, {X @ x: ~x}))

    def test_id_star_line_3(self):
        """Check to see if the counterfactual event is tautological."""
        self.assertEqual({}, remove_event_tautologies({X @ x: x}))
        self.assertEqual({Y @ x: y}, remove_event_tautologies({Y @ x: y, X @ x: x}))
        self.assertEqual({Y @ x: +y}, remove_event_tautologies({Y @ x: +y, X @ ~x: ~x}))
        self.assertEqual({Y @ x: +y}, remove_event_tautologies({Y @ x: +y, X @ (~x, w): ~x}))
        self.assertEqual({Y @ x: +y}, remove_event_tautologies({Y @ x: +y, X @ (w, ~x): ~x}))
        event = {Y @ (+x, -z): +y, X: -x}
        self.assertEqual(event, remove_event_tautologies(event))

    def test_id_star_line_4(self):
        """Check that the counterfactual graph is correct."""
        new_graph, new_event = make_counterfactual_graph(
            graph=figure_9a.graph, event={Y @ -x: -y, X: +x, Z @ -d: -z, D: -d}
        )
        self.assert_graph_equal(figure_9c.graph, new_graph)
        self.assertEqual({Y @ x: y, X: +x, Z: -z, D: -d}, new_event)

        actual_graph3, actual_event3 = make_counterfactual_graph(
            graph=NxMixedGraph.from_edges(directed=[(D, Z), (Z, Y)]),
            event={Z @ -d: -z, Z: +z, D: -d},
        )

        # created a test case where id_star_line_4 returns None
        self.assertIsNone(actual_event3)

        event_4 = {Y @ (-x, -z): +y, X: +x}

        expected_graph = figure_9d.graph
        expected = expected_graph, event_4
        self.assertEqual(expected, make_counterfactual_graph(figure_9a.graph, event_4))
        self.assertFalse(figure_9d.graph.is_connected())

    def test_id_star_line_6(self):
        """For each district, intervene only on the Markov pillow of each district.

        This is the key difference between the original ID* and our ID*.
        Original ID* intervenes on all variables not in the district, but this creates problems with identifiability.
        The identifiability issues are resolved by intervening only on the Markov pillow of each district.
        """
        input_graph = tikka_figure_5.graph
        input_event = {Y @ -x: -y, X: +x, Z: -z, D: -d}
        expected_summand = {W}
        district1 = frozenset({X, Y @ -x})
        district2 = frozenset({Z})
        district3 = frozenset({W @ -x})
        district4 = frozenset({D})
        input_event = {Y @ -x: -y, X: +x, Z: -z, D: -d}
        expected_event1 = {Y @ (-z, -w): -y, X @ (-z, -w): +x}
        expected_event2 = {Z @ -d: -z}
        expected_event3 = {W @ -x: -w}
        expected_event4 = {D: -d}
        expected_markov_pillow_events = {
            district1: expected_event1,
            district2: expected_event2,
            district3: expected_event3,
            district4: expected_event4,
        }

        expected_district_interventions = {
            frozenset({X, Y @ -x}): {Y @ (-x, -z, -w, -d): -y, X @ (-z, -w, -d): +x},
            frozenset({Z}): {Z @ (-y, -x, -w, -d): -z},
            frozenset({W @ -x}): {W @ (-y, -x, -z, -d): -w},
            frozenset({D}): {D @ (-y, -x, -z, -w): -d},
        }
        # Create a counterfactual graph with at least 2 c-components and return the summand and interventions of each

        actual_summand, actual_district_events = id_star_line_6(input_graph, input_event)
        self.assertEqual(expected_summand, actual_summand)

        # Original ID* intervenes on all variables not in the district, but this creates problems with identifiability
        self.assertNotEqual(expected_district_interventions, actual_district_events)

        # The identifiability issues are resolved by intervening only on the Markov pillow.
        self.assertEqual(expected_markov_pillow_events, actual_district_events)

        # input_graph2 = figure_9c.graph
        # input_event = {Y @ ~x: ~y, X: -x, Z @ -d: -z, D: -d}

    def test_get_free_variables(self):
        """Test that each variable not in the event or self-intervened is marginalized out."""
        input_graph = tikka_figure_5.graph
        input_event = {Y @ -x: -y, X: +x, Z: -z, D: -d}
        expected_summand = {W}
        self.assertEqual(expected_summand, get_free_variables(input_graph, input_event))
        input_event2 = {Y @ -x: -y, X: +x, Z @ -d: -z, D: -d}
        expected_summand2 = {W}
        self.assertEqual(expected_summand2, get_free_variables(input_graph, input_event2))

    def test_get_conflicts(self):
        """Test conflicts between interventions in the graph and value assignments in the event."""
        input_graph = NxMixedGraph.from_edges(
            directed=[(X @ +x, Y @ +x), (X @ -x, Y @ -x)], undirected=[(Y @ +x, Y @ -x)]
        )
        event = {Y @ -x: -y, Y @ +x: +y}
        self.assertEqual([(-x, +x), (+x, -x)], sorted(get_conflicts(input_graph, event)))

    def test_get_evidence(self):
        """Test that the interventions and values of a event."""
        self.assertEqual({-x, -w, -d, -y}, get_evidence({Y @ -x: -y, W @ -d: -w}))
        self.assertEqual({-x, -w, -d, -y}, get_evidence({Y @ -x: -y, X: -x, W @ -d: -w}))

    def test_get_cf_interventions(self):
        """Test we extract all interventions from the CF variables and only the counterfactual variables."""
        self.assertEqual(set(), get_cf_interventions(NxMixedGraph.from_edges(directed=[(X, Y)])))
        self.assertEqual(
            {-x, -d, -w},
            get_cf_interventions(NxMixedGraph.from_edges(directed=[(X, Y @ (-x, -d, -w))])),
        )
        self.assertEqual(
            {-x, -d, -w},
            get_cf_interventions(
                NxMixedGraph.from_edges(directed=[(X @ (-d, -w), Y @ (-x, -d, -w))])
            ),
        )
        self.assertEqual(
            {-x, -d, -w},
            get_cf_interventions(NxMixedGraph.from_edges(directed=[(X @ (-d, -w), Y @ -x)])),
        )

    def test_get_events_of_district(self):
        """Ensure that each variable in the district is intervened on the Markov pillow."""
        input_graph = tikka_figure_5.graph
        input_event = {Y @ -x: -y, X: +x, Z: -z, D: -d}
        district1 = {X, Y @ -x}
        district2 = {Z}
        district3 = {W @ -x}
        district4 = {D}
        # districts = district1 | district2 | district3 | district4
        expected_event1 = {Y @ (-z, -w): -y, X @ (-z, -w): +x}
        expected_event2 = {Z @ -d: -z}
        expected_event3 = {W @ -x: -w}
        expected_event4 = {D: -d}
        self.assertEqual(
            expected_event1, get_events_of_district(input_graph, district1, input_event)
        )
        self.assertEqual(
            expected_event2, get_events_of_district(input_graph, district2, input_event)
        )
        self.assertEqual(
            expected_event3, get_events_of_district(input_graph, district3, input_event)
        )
        self.assertEqual(
            expected_event4, get_events_of_district(input_graph, district4, input_event)
        )
        input_event5 = {Y @ (-D, -X): -Y, X: +X, Z: -Z}
        graph5 = NxMixedGraph.from_edges(
            nodes=[Y @ (-D, -X), X, Z @ (-D, -X), Z, W @ (-D, -X), D @ (-D, -X), D, X @ (-D, -X)],
            directed=[
                (Z @ (-D, -X), Y @ (-D, -X)),
                (W @ (-D, -X), Y @ (-D, -X)),
                (D @ (-D, -X), Z @ (-D, -X)),
                (D, Z),
                (X @ (-D, -X), W @ (-D, -X)),
            ],
            undirected=[(Y @ (-D, -X), X), (Z @ (-D, -X), Z)],
        )
        district5 = {Z, Z @ (-D, -X)}
        expected_event5 = {Z @ -D: -z}
        self.assertEqual(expected_event5, get_events_of_district(graph5, district5, input_event5))
        input_graph_unidentifiable = tikka_unidentifiable_cfgraph.graph
        self.assertEqual(
            expected_event1,
            get_events_of_district(input_graph_unidentifiable, district1, input_event),
        )

    def test_get_markov_pillow(self):
        """Test that we correctly compute the Markov pillow of the district."""
        district = {Z, Z @ (-d, -x)}
        graph = NxMixedGraph.from_edges(
            nodes=[Y @ (-D, -X), X, Z @ (-D, -X), Z, W @ (-D, -X), D @ (-D, -X), D, X @ (-D, -X)],
            directed=[
                (Z @ (-D, -X), Y @ (-D, -X)),
                (W @ (-D, -X), Y @ (-D, -X)),
                (D @ (-D, -X), Z @ (-D, -X)),
                (D, Z),
                (X @ (-D, -X), W @ (-D, -X)),
            ],
            undirected=[(Y @ (-D, -X), X), (Z @ (-D, -X), Z)],
        )
        expected_pillow = {D, D @ (-d, -x)}
        self.assertEqual(expected_pillow, graph.get_markov_pillow(district))
        graph2 = tikka_unidentifiable_cfgraph.graph
        district2 = {X, Y @ (-x)}
        expected_pillow2 = {W @ -x, Z}
        self.assertEqual(expected_pillow2, graph2.get_markov_pillow(district2))

    def test_events_of_each_district(self):
        """Ensure that each variable in the district is intervened on the Markov pillow."""
        input_graph = tikka_figure_5.graph
        district1 = frozenset({X, Y @ -x})
        district2 = frozenset({Z})
        district3 = frozenset({W @ -x})
        district4 = frozenset({D})
        input_event = {Y @ -x: -y, X: +x, Z: -z, D: -d}
        expected_event1 = {Y @ (-z, -w): -y, X @ (-z, -w): +x}
        expected_event2 = {Z @ -d: -z}
        expected_event3 = {W @ -x: -w}
        expected_event4 = {D: -d}
        expected_events = {
            district1: expected_event1,
            district2: expected_event2,
            district3: expected_event3,
            district4: expected_event4,
        }
        self.assertEqual(expected_events, get_events_of_each_district(input_graph, input_event))

    def test_intervene_on_district(self):
        """Test that we correctly intervene on a district."""
        district1 = {X, Y @ -x}
        district2 = {Z}
        district3 = {W @ -x}
        district4 = {D}
        districts = district1 | district2 | district3 | district4
        intervention1 = districts - district1
        intervention2 = districts - district2
        intervention3 = districts - district3
        intervention4 = districts - district4

        input_event = {Y @ -x: -y, X: +x, Z: -z, D: -d}  # from gamma' (Tikka 2022)
        expected_event1 = {Y @ (-x, -z, -w, -d): -y, X @ (-z, -w, -d): +x}
        expected_event2 = {Z @ (-y, -x, -w, -d): -z}
        expected_event3 = {W @ (-y, -x, -z, -d): -w}
        expected_event4 = {D @ (-y, -x, -z, -w): -d}
        self.assertEqual(
            expected_event1, intervene_on_district(district1, intervention1, input_event)
        )
        self.assertEqual(
            expected_event2, intervene_on_district(district2, intervention2, input_event)
        )
        self.assertEqual(
            expected_event3, intervene_on_district(district3, intervention3, input_event)
        )
        self.assertEqual(
            expected_event4, intervene_on_district(district4, intervention4, input_event)
        )

    def test_get_district_interventions(self):
        """Ensure that for each district, we intervene on the domain of each variable not in the district.

        Confirm that the domain of variables in the event query are restricted to their event value
        """
        # counterfactual_graph = NxMixedGraph.from_edges(
        #     undirected=[(Y @ (~X, Z), X)],
        #     directed=[
        #         (W @ (~X, Z), Y @ (~X, Z)),
        #     ],
        # )

        # event = {Y @ (+X, -Z): -Y, X: -X}
        input_event = {Y @ -x: -y, X: +x, Z: -z, D: -d}  # from gamma' (Tikka 2022)
        input_graph = tikka_figure_5.graph

        # expected_summand = {W}
        expected_district_interventions = {
            frozenset({X, Y @ -x}): {Y @ (-x, -z, -w, -d): -y, X @ (-z, -w, -d): +x},
            frozenset({Z}): {Z @ (-y, -x, -w, -d): -z},
            frozenset({W @ -x}): {W @ (-y, -x, -z, -d): -w},
            frozenset({D}): {D @ (-y, -x, -z, -w): -d},
        }
        actual = get_district_interventions(input_graph, input_event)
        self.assertEqual(expected_district_interventions, actual)

    def test_id_star_line_8(self):
        """Attempt to generate a conflict with an inconsistent value assignment."""
        input_graph = NxMixedGraph.from_edges(
            directed=[(X @ +x, Y @ +x), (X @ -x, Y @ -x)], undirected=[(Y @ +x, Y @ -x)]
        )
        event = {Y @ -x: -y, Y @ +x: +y}
        self.assertTrue(get_conflicts(input_graph, event))

    def test_id_star_line_9(self):
        """Test line 9 of the ID* algorithm.

        Test that estimand returned by taking the effect of all subscripts in
        new_event on variables in new_event is correct
        """
        input_graph1 = NxMixedGraph.from_edges(undirected=[(X, Y @ (-w, +x, -z))])
        expected1 = P[-w, +x, -z](X, Y)
        input_graph2 = NxMixedGraph.from_edges(undirected=[(X, Y)])
        expected2 = P(X, Y)
        self.assertEqual(expected1, id_star_line_9(input_graph1))
        self.assertEqual(expected2, id_star_line_9(input_graph2))

    def test_is_redundant_counterfactual(self):
        """Test that we can detect if counterfactual variable is redundant."""
        self.assertTrue(is_redundant_counterfactual(Y @ (+x, -y), -y))
        self.assertFalse(is_redundant_counterfactual(Y @ (+x, -y), +y))
        self.assertFalse(is_redundant_counterfactual(Y @ (+x, -z), -y))
        self.assertFalse(is_redundant_counterfactual(Y @ (+x, -z), +y))

    def test_get_remaining_and_missing_events(self):
        """Test that we can extract the missing and remaining events from a new event."""
        old_event = {Y @ -X: -Y, X: +X, Z @ -D: -Z, D: -D}
        new_event = {Y @ -X: -Y, X: +X, D: -D, Z: -Z}
        remaining_events, missing_events = get_remaining_and_missing_events(new_event, old_event)
        expected_remaining = {Y @ -x: -y, X: +x, D: -d}
        expected_missing = {Z @ -d: -z}
        self.assertEqual(expected_remaining, remaining_events)
        self.assertEqual(expected_missing, missing_events)

    def test_get_new_outcomes_and_conditions(self):
        """Test that we can recover all cases of outcomes and conditions from a new event."""
        old_outcomes = {Y @ -x: -y}
        old_conditions = {X: +x, Z @ -d: -z, D: -d}
        new_event = {Y @ -X: -Y, X: +X, D: -D, Z: -Z}
        new_outcomes, new_conditions = get_new_outcomes_and_conditions(
            new_event, old_outcomes, old_conditions
        )
        expected_new_outcomes = {Y @ -x: -y}
        expected_new_conditions = {X: +x, D: -d, Z: -z}
        self.assertEqual(expected_new_outcomes, new_outcomes)
        self.assertEqual(expected_new_conditions, new_conditions)
        new_event2 = {Y: -y, X: +x, D: -d, Z: -z}
        new_outcomes2, new_conditions2 = get_new_outcomes_and_conditions(
            new_event2, old_outcomes, old_conditions
        )
        expected_new_outcomes2 = {Y: -y}
        expected_new_conditions2 = {X: +x, D: -d, Z: -z}
        self.assertEqual(expected_new_outcomes2, new_outcomes2)
        self.assertEqual(expected_new_conditions2, new_conditions2)
        new_event3 = {Y: -y, X: +x, D: -d, Z @ -d: -z}
        new_outcomes3, new_conditions3 = get_new_outcomes_and_conditions(
            new_event3, old_outcomes, old_conditions
        )
        expected_new_outcomes3 = {Y: -y}
        expected_new_conditions3 = {X: +x, D: -d, Z @ -d: -z}
        self.assertEqual(expected_new_outcomes3, new_outcomes3)
        self.assertEqual(expected_new_conditions3, new_conditions3)
        new_event4 = {Y @ -x: -y, X: +x, D: -d, Z @ -d: -z}
        new_outcomes4, new_conditions4 = get_new_outcomes_and_conditions(
            new_event4, old_outcomes, old_conditions
        )
        expected_new_outcomes4 = {Y @ -x: -y}
        expected_new_conditions4 = {X: +x, D: -d, Z @ -d: -z}
        self.assertEqual(expected_new_outcomes4, new_outcomes4)
        self.assertEqual(expected_new_conditions4, new_conditions4)

        expected_new_outcomes5 = {Y: -y}
        expected_new_conditions5 = {Y: -y, Y @ -z: +y}
        new_event5 = {Y: -y, Y @ -z: +y}
        new_outcomes5, new_conditions5 = get_new_outcomes_and_conditions(
            new_event5, outcomes={Y @ -w: -y}, conditions={Y @ -x: -y, Y @ -z: +y}
        )
        self.assertEqual(expected_new_outcomes5, new_outcomes5)
        self.assertEqual(expected_new_conditions5, new_conditions5)

    def test_id_star(self):
        """Test that the ID* algorithm returns the correct estimand."""
        query = {Y @ (+x, -z): +y, X: -x}
        actual = id_star(figure_9a.graph, query)
        expected = Sum[W](P(Y @ (-z, -w), X @ (-z, -w)) * P(W @ -x))
        self.assert_expr_equal(expected, actual)

        input_graph2 = figure_9a.graph
        input_event2 = {Y @ -x: -y, X: +x, Z @ -d: -z, D: -d}
        expected2 = Sum[W](P(Y @ (-z, -w), X @ (-z, -w)) * P(Z @ -d) * P(W @ -x) * P(D))
        self.assert_expr_equal(expected2, id_star(input_graph2, input_event2))

        input_graph_line3 = input_graph2
        input_event_line3 = {D @ +d: +d}
        self.assert_expr_equal(One(), id_star(input_graph_line3, input_event_line3))

        input_graph_line5 = NxMixedGraph.from_edges(directed=[(D, Z), (Z, Y)])
        input_event_line5 = {Z @ -d: -z, Z: +z, D: -d}
        self.assert_expr_equal(Zero(), id_star(input_graph_line5, input_event_line5))

        input_graph_line8 = NxMixedGraph.from_edges(directed=[(X, Y)])
        with self.assertRaises(ConflictUnidentifiable):
            id_star(input_graph_line8, {Y @ -x: -y, Y @ +x: +y})

    def test_idc_star(self):
        """Test that the IDC* algorithm returns the correct estimand."""
        input_graph_line_1 = NxMixedGraph.from_edges(directed=[(D, Z), (Z, Y)])
        input_condition_line_1 = {Z @ -d: -z, Z: +z, D: -d}
        input_outcome_line_1 = {Y @ -d: -y}
        with self.assertRaises(ValueError):
            idc_star(input_graph_line_1, input_outcome_line_1, input_condition_line_1)

        input_graph_line3 = NxMixedGraph.from_edges(directed=[(D, Z), (Z, Y)])
        input_outcome_line3 = {Z @ -d: -z}
        input_condition_line3 = {Z: +z, D: -d}
        self.assert_expr_equal(
            Zero(), idc_star(input_graph_line3, input_outcome_line3, input_condition_line3)
        )

        input_graph_figure2 = tikka_figure_2.graph
        input_outcome_figure2 = {Y @ -x: -y}
        input_conditional_figure2 = {Z @ -x: -z, X: +x}
        expected_estimand_figure2 = P(Y @ (-x, -z))
        self.assert_expr_equal(
            expected_estimand_figure2,
            idc_star(input_graph_figure2, input_outcome_figure2, input_conditional_figure2),
        )

        input_graph_figure9a = figure_9a.graph
        input_outcome_figure9a = {Y @ -x: -y}
        input_conditional_figure9a = {X: +x, Z @ -d: -z, D: -d}
        id_star_estimand = Sum[W](P[-z, -w](-y, +x) * P[-x](-w))
        idc_star_estimand = id_star_estimand / Sum[W, Y, Z](P[-z, -w](-y, +x) * P[-x](-w))
        self.assert_expr_equal(
            idc_star_estimand, id_star_estimand / Sum[W, Y, Z](P[-z, -w](-y, +x) * P[-x](-w))
        )
        expected_estimand_figure_9a = Sum[D, W](
            P(Z @ -D) * P(X @ (-W, -Z), Y @ (-W, -Z)) * P(W @ -X) * P(D)
        ) / Sum[D, W, Y](Sum[D, W](P(Z @ -D) * P(X @ (-W, -Z), Y @ (-W, -Z)) * P(W @ -X) * P(D)))

        self.assert_expr_equal(
            expected_estimand_figure_9a,
            idc_star(input_graph_figure9a, input_outcome_figure9a, input_conditional_figure9a),
        )



# -*- coding: utf-8 -*-

"""Tests for canceling fractions."""

import unittest

from y0.dsl import A, B, Fraction, One, P, Sum

one = One()


class TestCancel(unittest.TestCase):
    """Test cancelling fractions."""

    def test_simple_identity(self):
        """Test cancelling when the numerator and denominator are the same."""
        for label, frac in [
            ("one", Fraction(one, one)),
            ("prob", P(A) / P(A)),
            ("sum", Sum.safe(P(A), ranges=[B]) / Sum.safe(P(A), ranges=[B])),
            ("product", (P(A) * P(B)) / (P(A) * P(B))),
        ]:
            with self.subTest(type=label):
                self.assertIsInstance(frac, Fraction)
                self.assertEqual(one, frac.simplify(), msg=f"\n\nActual:{frac}")

    def test_fraction_simplify(self):
        """Test cancelling on products."""
        for label, expected, frac in [
            ("leave num.", P(B), (P(A) * P(B)) / P(A)),
            ("leave den.", one / P(B), P(A) / (P(A) * P(B))),
            ("unordered", one, (P(A) * P(B)) / (P(B) * P(A))),
            ("canonical", one / P(A), one / P(A)),
            ("flipper", P(A), Fraction(one, Fraction(one, P(A)))),
            ("prob-redundant-one", P(A), Fraction(P(A), one)),
            ("sum-redundant-one", Sum[B](P(A)), Fraction(Sum[B](P(A)), one)),
            ("frac-redundant-one", P(A) / P(B), Fraction(Fraction(P(A), P(B)), one)),
            ("prod-redundant-one", P(A) * P(B), Fraction(P(A) * P(B), one)),
        ]:
            with self.subTest(type=label):
                self.assertIsInstance(frac, Fraction)
                self.assertEqual(expected, frac.simplify(), msg=f"\n\nActual:{frac}")



"""Unit tests for transport."""

import unittest

from y0.algorithm.transport import (
    TransportQuery,
    TRSOQuery,
    activate_domain_and_interventions,
    create_transport_diagram,
    get_nodes_to_transport,
    identify_target_outcomes,
    surrogate_to_transport,
    transport_variable,
    trso,
    trso_line1,
    trso_line2,
    trso_line3,
    trso_line4,
    trso_line6,
    trso_line9,
    trso_line10,
)
from y0.dsl import (
    PP,
    TARGET_DOMAIN,
    X1,
    X2,
    Y1,
    Y2,
    Expression,
    Fraction,
    Pi1,
    Pi2,
    Pi3,
    PopulationProbability,
    Probability,
    Product,
    Sum,
    Variable,
    W,
    X,
    Y,
    Z,
    Zero,
)
from y0.examples import tikka_trso_figure_8_graph as tikka_trso_figure_8
from y0.graph import NxMixedGraph
from y0.mutate import canonicalize, fraction_expand

tikka_trso_figure_8_transport = {
    X1: [Pi1],
    Y2: [Pi1],
    X2: [Pi2],
}

graph_1 = NxMixedGraph.from_edges(
    undirected=[(X1, Y1), (Z, W), (Z, X2)],
    directed=[
        (X1, Y1),
        (X1, Y2),
        (W, Y1),
        (W, Y2),
        (Z, Y1),
        (Z, X2),
        (X2, Y2),
        (Z, Y2),
        (transport_variable(X1), X1),
        (transport_variable(Y2), Y2),
    ],
)
graph_2 = NxMixedGraph.from_edges(
    undirected=[(X1, Y1), (Z, W), (Z, X2)],
    directed=[
        (X1, Y1),
        (X1, Y2),
        (W, Y1),
        (W, Y2),
        (Z, Y1),
        (Z, X2),
        (X2, Y2),
        (Z, Y2),
        (transport_variable(X2), X2),
    ],
)


class _TestCase(unittest.TestCase):
    def assert_expr_equal(self, expected: Expression, actual: Expression) -> None:
        """Assert that two expressions are the same."""
        ordering = tuple(expected.get_variables())
        expected_canonical = canonicalize(expected, ordering)
        actual_canonical = canonicalize(actual, ordering)
        self.assertEqual(
            expected_canonical,
            actual_canonical,
            msg=f"\nExpected: {str(expected_canonical)}\nActual:   {str(actual_canonical)}",
        )


class TestTransport(_TestCase):
    """Test surrogate outcomes and transportability."""

    def test_create_transport_diagram(self):
        """Test that we can create the transport diagram correctly."""
        graph_pi1 = create_transport_diagram(graph=tikka_trso_figure_8, nodes_to_transport={X1, Y2})
        graph_pi2 = create_transport_diagram(graph=tikka_trso_figure_8, nodes_to_transport={X2})

        self.assertEqual(graph_1, graph_pi1)
        self.assertEqual(graph_2, graph_pi2)

    def test_get_nodes_to_transport(self):
        """Test that we can correctly find the nodes which should have transport nodes."""
        expected = {X1, Y2}
        actual = get_nodes_to_transport(
            surrogate_interventions=X1, surrogate_outcomes=Y1, graph=tikka_trso_figure_8
        )
        self.assertEqual(actual, expected)
        expected = {X2}
        actual = get_nodes_to_transport(
            surrogate_interventions={X2}, surrogate_outcomes={Y2}, graph=tikka_trso_figure_8
        )
        self.assertEqual(actual, expected)

        # Test for multiple vertices in interventions and surrogate_outcomes
        expected = {X1, X2, Y1}
        actual = get_nodes_to_transport(
            surrogate_interventions={X2, X1}, surrogate_outcomes={Y2, W}, graph=tikka_trso_figure_8
        )
        self.assertEqual(actual, expected)

    def test_surrogate_to_transport(self):
        """Test that surrogate_to_transport correctly converts to a transport query."""
        target_outcomes = {Y1, Y2}
        target_interventions = {X1, X2}
        surrogate_outcomes = {Pi1: {Y1}, Pi2: {Y2}}
        surrogate_interventions = {Pi1: {X1}, Pi2: {X2}}

        actual = surrogate_to_transport(
            target_outcomes=target_outcomes,
            target_interventions=target_interventions,
            graph=tikka_trso_figure_8,
            surrogate_outcomes=surrogate_outcomes,
            surrogate_interventions=surrogate_interventions,
        )

        expected = TransportQuery(
            target_interventions=target_interventions,
            target_outcomes=target_outcomes,
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            domains={Pi1, Pi2},
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
            target_experiments=set(),
        )
        self.assertEqual(actual, expected)
        extra_surrogate_outcomes = {Pi1: {Y1}, Pi2: {Y2}, Pi3: {Y2}}
        missing_surrogate_interventions = {Pi1: {Y2}}
        with self.assertRaises(ValueError):
            surrogate_to_transport(
                target_outcomes=target_outcomes,
                target_interventions=target_interventions,
                graph=tikka_trso_figure_8,
                surrogate_outcomes=extra_surrogate_outcomes,
                surrogate_interventions=surrogate_interventions,
            )
        with self.assertRaises(ValueError):
            surrogate_to_transport(
                target_outcomes=target_outcomes,
                target_interventions=target_interventions,
                graph=tikka_trso_figure_8,
                surrogate_outcomes=surrogate_outcomes,
                surrogate_interventions=missing_surrogate_interventions,
            )

    def test_add_active_interventions(self):
        """Test that interventions are added correctly."""
        expected_1 = fraction_expand(PP[Pi1][X1](Y1 | W, Z))
        expected_2 = fraction_expand(PP[Pi2][X2](Y2 | W, Z, X1))
        test_1 = fraction_expand(PP[Pi1](Y1 | W, Z))
        test_2 = fraction_expand(PP[Pi2](Y2 | W, Z, X1))
        # expected = canonicalize(
        #     Sum.safe(
        #         Product.safe([expected_part1, expected_part2, expected_part3]),
        #         (W, Z),
        #     )
        # )
        # expected = Sum[Y1](PP[Pi1][X1](W, Y1, Z)
        actual_1 = activate_domain_and_interventions(test_1, {X1}, Pi1)
        self.assert_expr_equal(expected_1, actual_1)
        actual_2 = activate_domain_and_interventions(test_2, {X2}, Pi2)
        self.assert_expr_equal(expected_2, actual_2)

        expected_3 = Sum.safe(expected_1, (W, Z))
        test_3 = Sum.safe(test_1, (W, Z))
        actual_3 = activate_domain_and_interventions(test_3, {X1}, Pi1)
        self.assert_expr_equal(expected_3, actual_3)

        # TODO fix this test
        # expected_1b = fraction_expand(PP[Pi1][X2](Y1 | W, Z))
        # expected_4 = Product.safe((expected_1b, expected_2))
        # test_4 = Product.safe((test_1, test_2))
        # actual_4 = activate_domain_and_interventions(test_4, X2)
        # self.assert_expr_equal(expected_4, actual_4)

    def test_trso_line1(self):
        """Test that trso_line 1 returns the correct expression."""
        # triggers line 1
        outcomes = {Y1, Y2}
        domain_graph = tikka_trso_figure_8
        expression = PP[TARGET_DOMAIN](domain_graph.nodes())

        expected = Sum.safe(expression, {W, X1, X2, Z})
        actual = trso_line1(
            outcomes,
            expression,
            domain_graph,
        )
        self.assert_expr_equal(expected, actual)

    def test_trso_line2(self):
        """Test that trso _line2 correctly modifies the query."""
        query_part1 = TRSOQuery(
            target_interventions={X1, X2, Y1, Y2},
            target_outcomes={Z, W},
            expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        # actual_part1 = trso(query_part1)
        # triggers line 2
        graph = query_part1.graphs[query_part1.domain]
        outcome_ancestors = graph.ancestors_inclusive(query_part1.target_outcomes)
        actual_part1 = trso_line2(query_part1, outcome_ancestors)

        outcomes_anc = {W, Z}
        outcomes_anc_pi1 = graph_1.ancestors_inclusive(query_part1.target_outcomes)
        outcomes_anc_pi2 = graph_2.ancestors_inclusive(query_part1.target_outcomes)
        new_query_part1 = TRSOQuery(
            target_interventions=set(),
            target_outcomes={W, Z},
            expression=canonicalize(PP[TARGET_DOMAIN]((W, Z))),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph(outcomes_anc),
                Pi1: graph_1.subgraph(outcomes_anc_pi1),
                Pi2: graph_2.subgraph(outcomes_anc_pi2),
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        expected_part1 = new_query_part1
        self.assertEqual(actual_part1, expected_part1)

        # this is the simplified form of the expression
        # Maybe to be done in some future implementation
        # expected3 = PP[TARGET_DOMAIN]({W,Z})
        # actual_part1_simplified = actual_part1.simplified
        # self.assertEqual(expected3, actual_part1_simplified)

        query_part2 = TRSOQuery(
            target_interventions={X1, X2, Z, W, Y2},
            target_outcomes={Y1},
            expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        # actual_part2 = trso(query_part2)
        # should trigger line 2
        graph = query_part2.graphs[query_part2.domain]
        outcome_ancestors = graph.ancestors_inclusive(query_part2.target_outcomes)
        actual_part2 = trso_line2(query_part2, outcome_ancestors)
        # triggers line 2

        outcomes_anc = {X1, W, Z, Y1}
        outcomes_anc_pi1 = graph_1.ancestors_inclusive(query_part2.target_outcomes)
        outcomes_anc_pi2 = graph_2.ancestors_inclusive(query_part2.target_outcomes)

        new_query_part2 = TRSOQuery(
            target_interventions={X1, W, Z},
            target_outcomes={Y1},
            expression=canonicalize(PP[TARGET_DOMAIN](X1, W, Z, Y1)),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph(outcomes_anc),
                Pi1: graph_1.subgraph(outcomes_anc_pi1),
                Pi2: graph_2.subgraph(outcomes_anc_pi2),
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        expected_part2 = new_query_part2
        self.assertEqual(actual_part2, expected_part2)

        query_3 = TRSOQuery(
            target_interventions={X1, X2, Z, W, Y2},
            target_outcomes={Y1},
            expression=Sum.safe(PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()), (W, Z)),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        graph = query_3.graphs[query_3.domain]
        outcome_ancestors = graph.ancestors_inclusive(query_3.target_outcomes)
        actual_3 = trso_line2(query_3, outcome_ancestors)

        outcomes_anc = {X1, W, Z, Y1}
        outcomes_anc_pi1 = graph_1.ancestors_inclusive(query_3.target_outcomes)
        outcomes_anc_pi2 = graph_2.ancestors_inclusive(query_3.target_outcomes)
        expected_3 = TRSOQuery(
            target_interventions={X1, W, Z},
            target_outcomes={Y1},
            expression=Sum.safe(
                Sum.safe(PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()), (W, Z)), (X2, Y2)
            ),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph(outcomes_anc),
                Pi1: graph_1.subgraph(outcomes_anc_pi1),
                Pi2: graph_2.subgraph(outcomes_anc_pi2),
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        self.assertEqual(actual_3, expected_3)

    def test_trso_line3(self):
        """Test that trso_line3 correctly modifies the query."""
        transportability_diagram_line3 = NxMixedGraph.from_edges(
            directed=[
                (W, X),
                (X, Y),
                (X, Z),
                (Z, Y),
            ],
        )

        query = TRSOQuery(
            target_interventions={X},
            target_outcomes={Y},
            expression=PP[TARGET_DOMAIN](transportability_diagram_line3.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: transportability_diagram_line3,
            },
            surrogate_interventions={},
        )

        # target_interventions_overbar = transportability_diagram_line3.remove_in_edges(
        #     query.target_interventions
        # )
        # additional_interventions = (
        #     transportability_diagram_line3.nodes()
        #     - query.target_interventions
        #     - target_interventions_overbar.ancestors_inclusive(query.target_outcomes)
        # )
        additional_interventions = {W}

        new_query = TRSOQuery(
            target_interventions={X, W},
            target_outcomes={Y},
            expression=PP[TARGET_DOMAIN](transportability_diagram_line3.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: transportability_diagram_line3,
            },
            surrogate_interventions={},
        )
        expected = new_query

        actual = trso_line3(query=query, additional_interventions=additional_interventions)
        self.assertEqual(actual, expected)

        self.assertIsNotNone(trso(query))
        self.assertIsNotNone(trso(actual))

    def test_trso_line4(self):
        """Test that trso_line4 builds a dictionary of components and modified queries."""
        query = TRSOQuery(
            target_interventions={X1, X2},
            target_outcomes={Y1, Y2},
            expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        districts_without_interventions = tikka_trso_figure_8.remove_nodes_from(
            query.target_interventions
        ).districts()

        expected = {
            frozenset([Y2]): TRSOQuery(
                target_interventions={X1, X2, Z, W, Y1},
                target_outcomes={Y2},
                expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
                active_interventions=set(),
                domain=TARGET_DOMAIN,
                domains={Pi1, Pi2},
                graphs={
                    TARGET_DOMAIN: tikka_trso_figure_8,
                },
                surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
            ),
            frozenset([Y1]): TRSOQuery(
                target_interventions={X1, X2, Z, W, Y2},
                target_outcomes={Y1},
                expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
                active_interventions=set(),
                domain=TARGET_DOMAIN,
                domains={Pi1, Pi2},
                graphs={
                    TARGET_DOMAIN: tikka_trso_figure_8,
                },
                surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
            ),
            frozenset([W, Z]): TRSOQuery(
                target_interventions={X1, X2, Y1, Y2},
                target_outcomes={W, Z},
                expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
                active_interventions=set(),
                domain=TARGET_DOMAIN,
                domains={Pi1, Pi2},
                graphs={
                    TARGET_DOMAIN: tikka_trso_figure_8,
                },
                surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
            ),
        }

        actual = trso_line4(
            query,
            districts_without_interventions,
        )

        self.assertEqual(expected, actual)

    def test_trso_line6(self):
        """Test that trso_line6 builds a dictionary of domains and modified queries."""
        query = TRSOQuery(
            target_interventions={X1, Z, W},
            target_outcomes={Y1},
            expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        actual = trso_line6(
            query,
        )
        new_transportability_diagram = query.graphs[Pi1].subgraph(
            query.graphs[Pi1].nodes()
            - query.surrogate_interventions[Pi1].intersection(query.target_interventions)
        )
        expected = {
            Pi1: TRSOQuery(
                target_interventions={Z, W},
                target_outcomes={Y1},
                expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
                active_interventions={X1},
                domain=Pi1,
                domains={Pi1, Pi2},
                graphs={
                    TARGET_DOMAIN: tikka_trso_figure_8,
                    Pi1: new_transportability_diagram,
                    Pi2: graph_2,
                },
                surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
            )
        }
        self.assertEqual(expected, actual)

        outcomes_anc = {X1, W, Z, Y1}
        query_part2 = TRSOQuery(
            target_interventions={X1, W, Z},
            target_outcomes={Y1},
            expression=Sum.safe(PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()), {X2, Y2}),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph(outcomes_anc),
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        actual_part2 = trso_line6(
            query_part2,
        )

        new_transportability_diagram = query_part2.graphs[Pi1].subgraph(
            query_part2.graphs[Pi1].nodes()
            - query_part2.surrogate_interventions[Pi1].intersection(
                query_part2.target_interventions
            )
        )
        expected_part2 = {
            Pi1: TRSOQuery(
                target_interventions={Z, W},
                target_outcomes={Y1},
                expression=Sum.safe(PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()), {X2, Y2}),
                active_interventions={X1},
                domain=Pi1,
                domains={Pi1, Pi2},
                graphs={
                    TARGET_DOMAIN: tikka_trso_figure_8.subgraph(outcomes_anc),
                    Pi1: new_transportability_diagram,
                    Pi2: graph_2,
                },
                surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
            )
        }
        self.assertEqual(expected_part2, actual_part2)

    def test_trso_line9(self):
        """Test that trso_line3 returns the correct expression."""
        line9_query1 = TRSOQuery(
            target_interventions={W, Z, X1},
            target_outcomes={Y2},
            expression=PP[Pi2](W, X1, Y2, Z),
            active_interventions={X2},
            domain=Pi2,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph({W, X1, Y2, Z}),
                Pi1: graph_1.subgraph({W, X1, Y2, Z}),
                Pi2: graph_2.subgraph({W, X1, Y2, Z}),
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        district1 = {Y2}
        line9_actual1 = trso_line9(line9_query1, district1)
        line9_expected1 = PP[Pi2](W, X1, Y2, Z) / Sum.safe(PP[Pi2](W, X1, Y2, Z), (Y2,))
        self.assertEqual(line9_expected1, line9_actual1)

        line9_query2 = TRSOQuery(
            target_interventions={W, Z},
            target_outcomes={Y1},
            expression=PP[Pi2](W, Y1, Z),
            active_interventions={X1},
            domain=Pi1,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph({W, Y1, Z}),
                Pi1: graph_1.subgraph({W, Y1, Z}),
                Pi2: graph_2.subgraph({W, Y1, Z}),
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        district2 = {Y1}
        line9_actual2 = trso_line9(line9_query2, district2)
        line9_expected2 = PP[Pi2](W, Y1, Z) / Sum.safe(PP[Pi2](W, Y1, Z), (Y1,))
        self.assertEqual(line9_expected2, line9_actual2)

        line9_query3 = TRSOQuery(
            target_interventions={W, Z},
            target_outcomes={Y1},
            expression=Zero(),
            active_interventions={X1},
            domain=Pi1,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph({W, Y1, Z}),
                Pi1: graph_1.subgraph({W, Y1, Z}),
                Pi2: graph_2.subgraph({W, Y1, Z}),
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        district3 = {Y1}
        with self.assertRaises(RuntimeError):
            trso_line9(line9_query3, district3)

    def test_trso_line10(self):
        """Test that trso_line10 correctly modifies the query."""
        # pass
        line10_query1 = TRSOQuery(
            target_interventions={W, Z, X1},
            target_outcomes={Y2},
            expression=PP[Pi2](W, X1, Y2, Z),
            active_interventions={X2},
            domain=Pi2,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph({W, X1, Y2, Z}),
                Pi1: graph_1.subgraph({W, X1, Y2, Z}),
                Pi2: graph_2.subgraph({W, X1, Y2, Z}),
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        district1 = {Y2}
        new_surrogate_interventions = dict()
        line10_actual1 = trso_line10(line10_query1, district1, new_surrogate_interventions)

        ordering = list(line10_query1.graphs[Pi2].topological_sort())
        i = ordering.index(Y2)
        pre_node = set(ordering[:i])
        prob = Probability.safe(Y2 | pre_node)
        new_expression = PopulationProbability(population=Pi2, distribution=prob.distribution)

        line10_expected1 = TRSOQuery(
            target_interventions=set(),
            target_outcomes={Y2},
            expression=new_expression,
            active_interventions={X2},
            domain=Pi2,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8.subgraph({W, X1, Y2, Z}),
                Pi1: graph_1.subgraph({W, X1, Y2, Z}),
                Pi2: graph_2.subgraph({Y2}),
            },
            surrogate_interventions=dict(),
        )

        self.assertEqual(line10_expected1, line10_actual1)


class TestIntegration(_TestCase):
    """Test integration over the whole workflow."""

    def test_transport_variable(self):
        """Test that transport nodes will not take inappropriate inputs."""
        with self.assertRaises(TypeError):
            transport_variable(Y1 @ -X2)

    def test_trso_1(self):
        """Test that trso returns the correct expression."""
        query_part1 = TRSOQuery(
            target_interventions={X1, X2, Y1, Y2},
            target_outcomes={Z, W},
            expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        actual_part1 = trso(query_part1)

        graph = query_part1.graphs[query_part1.domain]
        outcome_ancestors = graph.ancestors_inclusive(query_part1.target_outcomes)
        new_query_part1 = trso_line2(query_part1, outcome_ancestors)
        new_graph = new_query_part1.graphs[new_query_part1.domain]
        expected_part1 = canonicalize(
            trso_line1(new_query_part1.target_outcomes, new_query_part1.expression, new_graph)
        )
        self.assertEqual(expected_part1, actual_part1)
        expected2 = canonicalize(PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes() - {X1, X2, Y1, Y2}))
        self.assertEqual(expected2, actual_part1)

        # this is the simplified form of the expression
        # Maybe to be done in some future implementation
        # expected3 = PP[TARGET_DOMAIN]({W,Z})
        # actual_part1_simplified = actual_part1.simplified
        # self.assertEqual(expected3, actual_part1_simplified)

    def test_trso_2(self):
        """Test that trso returns the correct expression."""
        query_part2 = TRSOQuery(
            target_interventions={X1, X2, Z, W, Y2},
            target_outcomes={Y1},
            expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        actual_part2 = trso(query_part2)
        expected_part2_conditional = canonicalize(
            PP[Pi1](Y1 @ X1, Z @ X1, W @ X1).conditional((Z @ X1, W @ X1))
        )
        expected_part2_magic_p = PP[Pi1][X1](Y1 | W, Z)
        expected_part2_full = PP[Pi1][X1](W, Y1, Z) / Sum[Y1](PP[Pi1][X1](W, Y1, Z))

        self.assertIsInstance(actual_part2, Fraction)
        self.assert_expr_equal(expected_part2_full.numerator, actual_part2.numerator)
        self.assert_expr_equal(expected_part2_full, actual_part2)
        self.assert_expr_equal(expected_part2_conditional, actual_part2)
        self.assert_expr_equal(fraction_expand(expected_part2_magic_p), actual_part2)

    def test_trso_3(self):
        """Test that trso returns the correct expression."""
        query_part3 = TRSOQuery(
            target_interventions={X1, X2, Z, W, Y1},
            target_outcomes={Y2},
            expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        actual_part3 = trso(query_part3)
        expected_part3_conditional = canonicalize(PP[Pi2][X2](Y2, X1, Z, W).conditional((X1, Z, W)))
        expected_part3_magic_p = PP[Pi2][X2](Y2 | W, Z, X1)
        expected_part3_full = PP[Pi2](W @ -X2, X1 @ -X2, Y2 @ -X2, Z @ -X2) / Sum[Y2](
            PP[Pi2](W @ -X2, X1 @ -X2, Y2 @ -X2, Z @ -X2)
        )
        self.assert_expr_equal(expected_part3_full, actual_part3)
        self.assert_expr_equal(expected_part3_conditional, actual_part3)
        self.assert_expr_equal(fraction_expand(expected_part3_magic_p), actual_part3)

    def test_trso_4(self):
        """Test that trso returns the correct expression."""
        query = TRSOQuery(
            target_interventions={X1, X2},
            target_outcomes={Y1, Y2},
            expression=PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes()),
            active_interventions=set(),
            domain=TARGET_DOMAIN,
            domains={Pi1, Pi2},
            graphs={
                TARGET_DOMAIN: tikka_trso_figure_8,
                Pi1: graph_1,
                Pi2: graph_2,
            },
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )

        actual = trso(query)
        expected_part1 = PP[TARGET_DOMAIN](W, Z)
        expected_part2 = PP[Pi1][X1](W, Y1, Z) / Sum[Y1](PP[Pi1][X1](W, Y1, Z))
        expected_part3 = PP[Pi2](W @ -X2, X1 @ -X2, Y2 @ -X2, Z @ -X2) / Sum[Y2](
            PP[Pi2](W @ -X2, X1 @ -X2, Y2 @ -X2, Z @ -X2)
        )

        expected = canonicalize(
            Sum.safe(
                Product.safe([expected_part1, expected_part2, expected_part3]),
                (W, Z),
            )
        )
        self.assert_expr_equal(expected, actual)

    def test_trso_5(self):
        """Test that trso returns the correct expression."""
        # TODO we should find a way to accomplish this with a transport call
        # This test triggers pillow_has_transport on line 10 and returns None
        target_interventions = {W, Z}
        target_outcomes = {Y1}
        surrogate_interventions = {Pi1: {X1}, Pi2: {X2}}
        surrogate_outcomes = {Pi1: {Y1}, Pi2: {Y2}}
        initial_expression = PP[TARGET_DOMAIN](tikka_trso_figure_8.nodes())
        transport_query = surrogate_to_transport(
            graph=tikka_trso_figure_8,
            target_outcomes=target_outcomes,
            target_interventions=target_interventions,
            surrogate_outcomes=surrogate_outcomes,
            surrogate_interventions=surrogate_interventions,
        )

        transport_z = transport_variable(Z)
        hacked_graph = NxMixedGraph.from_edges(
            undirected=[(X1, Y1), (Y1, W), (Z, X2), (Z, W)],
            directed=[
                (X1, Y1),
                (X1, Y2),
                (W, Y1),
                (W, Y2),
                (Z, Y1),
                (Z, W),
                (Z, X2),
                (X2, Y2),
                (Z, Y2),
                (transport_z, Z),
            ],
        )
        trso_query = TRSOQuery(
            target_interventions={Z, W},
            target_outcomes=transport_query.target_outcomes,
            expression=initial_expression,
            active_interventions={X1},
            domain=TARGET_DOMAIN,
            domains=transport_query.domains,
            graphs=transport_query.graphs,
            surrogate_interventions=transport_query.surrogate_interventions,
        )
        trso_query.graphs[trso_query.domain] = hacked_graph
        self.assertIsNone(trso(trso_query))

    def test_transport_1(self):
        """Test that transport returns the correct expression."""
        expected_part1 = PP[TARGET_DOMAIN](W, Z)
        self.assertIsInstance(expected_part1, PopulationProbability)
        expected_part2 = fraction_expand(PP[Pi1][X1](Y1 | W, Z))
        self.assertIsInstance(expected_part2, Fraction)
        self.assertIsInstance(expected_part2.numerator, PopulationProbability)
        self.assertIsInstance(expected_part2.denominator, PopulationProbability)
        expected_part3 = fraction_expand(PP[Pi2][X2](Y2 | W, Z, X1))
        expected_estimand = canonicalize(
            Sum.safe(
                Product.safe([expected_part1, expected_part2, expected_part3]),
                (W, Z),
            )
        )
        actual_estimand = identify_target_outcomes(
            graph=tikka_trso_figure_8,
            target_outcomes={Y1, Y2},
            target_interventions={X1, X2},
            surrogate_outcomes={Pi1: {Y1}, Pi2: {Y2}},
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        self.assert_expr_equal(expected_estimand, actual_estimand)

    def test_transport_2(self):
        """Test that transport returns the correct expression."""
        # This test triggers part of line 11 in trso (district length of 1)
        new_graph = tikka_trso_figure_8.subgraph(tikka_trso_figure_8.nodes() - {X1})
        new_graph.add_undirected_edge(W, Y1)
        actual_11 = identify_target_outcomes(
            graph=new_graph,
            target_outcomes={Y1},
            target_interventions={Z, W},
            surrogate_outcomes={Pi1: {Y1}, Pi2: {Y2}},
            surrogate_interventions={Pi1: {X2}, Pi2: {X2}},
        )
        self.assertIsNone(actual_11)

    def test_transport_3(self):
        """Test that transport returns the correct expression."""
        # This triggers triggers not implemented error on line 9
        # Now it triggers value error
        new_graph = tikka_trso_figure_8.subgraph(tikka_trso_figure_8.nodes() - {X1})
        with self.assertRaises(ValueError):
            identify_target_outcomes(
                graph=new_graph,
                target_outcomes={Z, Y1, W},
                target_interventions={Y1},
                surrogate_outcomes={Pi1: {Y1}, Pi2: {Y2}},
                surrogate_interventions={Pi1: {X2}, Pi2: {X2}},
            )

    def test_transport_4(self):
        """Test that transport returns the correct expression."""
        # This triggers line 10.
        # TODO it fails on the next recursive loop, would be better to find an example that doesn't fail.
        grpah = NxMixedGraph.from_edges(
            undirected=[(X1, Y1), (Y1, W), (Z, X2)],
            directed=[
                (X1, Y1),
                (X1, Y2),
                (W, Y1),
                (W, Y2),
                (Z, Y1),
                (Z, X2),
                (X2, Y2),
                (Z, Y2),
            ],
        )
        estimand = identify_target_outcomes(
            graph=grpah,
            target_outcomes={Y1},
            target_interventions={W, Z},
            surrogate_outcomes={Pi1: {Y1}, Pi2: {Y2}},
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        self.assertIsNone(estimand)

    def test_transport_5(self):
        """Test that transport returns the correct expression."""
        with self.assertRaises(ValueError):
            identify_target_outcomes(
                graph=tikka_trso_figure_8,
                target_outcomes={Y1},
                target_interventions={W, Z},
                surrogate_outcomes={Pi1: {Y1}, Pi2: {Y2}},
                surrogate_interventions={Pi1: {X1}, Pi2: {X2}, Pi3: {X}},
            )

    def test_transport_6(self):
        """Test that transport returns the correct expression."""
        # This test triggers if expression is None:  continue block after line 6
        graph = NxMixedGraph.from_edges(
            undirected=[(X1, Y1), (Z, W), (Z, X2), (Y2, X1)],
            directed=[
                (X1, Y1),
                (X1, Y2),
                (W, Y1),
                (W, Y2),
                (Z, Y1),
                (Z, X2),
                (X2, Y2),
                (Z, Y2),
            ],
        )
        estimand = identify_target_outcomes(
            graph,
            target_outcomes={Y1, Y2},
            target_interventions={X1, X2},
            surrogate_outcomes={Pi1: {Y1}, Pi2: {Y2}},
            surrogate_interventions={Pi1: {X1}, Pi2: {X2}},
        )
        self.assertIsNone(estimand)


class TestHighLevel(unittest.TestCase):
    """Tests on more high-level applications."""

    def test_grades(self):
        """Test grades scenario."""
        c, s, g, j = (Variable("C"), Variable("S"), Variable("G"), Variable("J"))
        graph = NxMixedGraph.from_edges(
            directed=[(c, s), (s, g), (g, j)],
            undirected=[(s, c), (j, c)],
        )
        estimand = identify_target_outcomes(
            graph=graph,
            target_outcomes={j},
            target_interventions={c},
            surrogate_outcomes={Pi1: {s}, Pi2: {j}},
            surrogate_interventions={Pi1: {c}, Pi2: {g}},
        )
        self.assertIsNotNone(estimand)



"""Utilities for generating examples."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional, Sequence

import pandas as pd

from y0.algorithm.identify import Identification, Query
from y0.dsl import Variable
from y0.graph import NxMixedGraph
from y0.struct import DSeparationJudgement, VermaConstraint

__all__ = [
    "Example",
]


@dataclass
class Example:
    """An example graph packaged with certain pre-calculated data structures."""

    name: str
    reference: str
    graph: NxMixedGraph
    description: Optional[str] = None
    verma_constraints: Optional[Sequence[VermaConstraint]] = None
    conditional_independencies: Optional[Sequence[DSeparationJudgement]] = None
    data: Optional[pd.DataFrame] = None
    identifications: Optional[list[dict[str, list[Identification]]]] = None
    #: Example queries are just to give an idea to a new user
    #: what might be interesting to use in the ID algorithm
    example_queries: Optional[list[Query]] = None
    generate_data: Optional[Callable[[int, Optional[dict[Variable, float]]], pd.DataFrame]] = None



# -*- coding: utf-8 -*-

"""Test graph construction and conversion."""

import unittest
from textwrap import dedent
from typing import Set, Tuple

import networkx as nx

from y0.dsl import A, B, C, D, M, Variable, X, Y, Z
from y0.examples import SARS_SMALL_GRAPH, Example, examples, napkin, verma_1
from y0.graph import (
    DEFAULT_TAG,
    DEFULT_PREFIX,
    NxMixedGraph,
    is_a_fixable,
    is_markov_blanket_shielded,
    is_p_fixable,
)
from y0.resources import VIRAL_PATHOGENESIS_PATH


class TestGraph(unittest.TestCase):
    """Test graph construction and conversion."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.addTypeEqualityFunc(NxMixedGraph, self.assert_graph_equal)

    def assert_graph_equal(self, a: NxMixedGraph, b: NxMixedGraph, msg=None) -> None:
        """Check the graphs are equal (more nice than the builtin :meth:`NxMixedGraph.__eq__` for testing)."""
        self.assertEqual(set(a.directed.nodes()), set(b.directed.nodes()), msg=msg)
        self.assertEqual(set(a.undirected.nodes()), set(b.undirected.nodes()), msg=msg)
        self.assertEqual(set(a.directed.edges()), set(b.directed.edges()), msg=msg)
        self.assertEqual(
            set(map(frozenset, a.undirected.edges())),
            set(map(frozenset, b.undirected.edges())),
            msg=msg,
        )

    def test_example_nodes(self):
        """Test all nodes are variables in example graphs."""
        for example in examples:
            with self.subTest(name=example.name):
                self.assertIsInstance(example, Example)
                non_variables = {
                    node for node in example.graph.nodes() if not isinstance(node, Variable)
                }
                self.assertEqual(0, len(non_variables), msg=f"Found non-variables: {non_variables}")

    def test_causaleffect_str_verma_1(self):
        """Test generating R code for the figure 1A graph for causaleffect."""
        expected = dedent(
            """
        g <- graph.formula(V1 -+ V2, V2 -+ V3, V3 -+ V4, V2 -+ V4, V4 -+ V2, simplify = FALSE)
        g <- set.edge.attribute(graph = g, name = "description", index = c(4, 5), value = "U")
        """
        ).strip()
        self.assertEqual(expected, verma_1.to_causaleffect_str())

    def assert_labeled_convertable(
        self, graph: NxMixedGraph, labeled_edges: Set[Tuple[str, str]]
    ) -> None:
        """Test that the graph can be converted to a DAG, then back to an ADMG."""
        prefix = DEFULT_PREFIX
        tag = DEFAULT_TAG

        labeled_dag = graph.to_latent_variable_dag(prefix=prefix, tag=tag)
        for node in labeled_dag:
            self.assertIn(tag, labeled_dag.nodes[node], msg=f"Node: {node}")
            self.assertEqual(node.startswith(prefix), labeled_dag.nodes[node][tag])

        self.assertEqual(labeled_edges, set(labeled_dag.edges()))

        reconstituted = NxMixedGraph.from_latent_variable_dag(labeled_dag, tag=tag)
        self.assertEqual(graph, reconstituted)

    def test_convertable(self):
        """Test graphs are convertable."""
        for graph, labeled_edges in [
            (
                verma_1,
                {
                    ("V1", "V2"),
                    ("V2", "V3"),
                    ("V3", "V4"),
                    (f"{DEFULT_PREFIX}0", "V2"),
                    (f"{DEFULT_PREFIX}0", "V4"),
                },
            ),
        ]:
            with self.subTest():
                self.assert_labeled_convertable(graph, labeled_edges)

    def test_from_causalfusion(self):
        """Test importing a CausalFusion graph."""
        graph = NxMixedGraph.from_causalfusion_path(VIRAL_PATHOGENESIS_PATH)
        self.assertIsInstance(graph, NxMixedGraph)

    def test_from_admg(self):
        """Test that all ADMGs can be converted to NxMixedGraph."""
        try:
            from ananke.graphs import ADMG
        except ImportError:
            self.skipTest("ananke is not available")

        expected = NxMixedGraph.from_str_adj(
            directed={"W": [], "X": ["Y"], "Y": ["Z"], "Z": []},
            undirected={"W": [], "X": ["Z"], "Y": [], "Z": []},
        )
        admg = ADMG(
            vertices=["W", "X", "Y", "Z"],
            di_edges=[["X", "Y"], ["Y", "Z"]],
            bi_edges=[["X", "Z"]],
        )
        self.assertEqual(expected, NxMixedGraph.from_admg(admg))

    def test_from_adj(self):
        """Test the adjacency graph is not a multigraph."""
        directed = dict([("a", ["b", "c"]), ("b", ["a"]), ("c", [])])
        expected = NxMixedGraph.from_str_edges(directed=[("a", "b"), ("a", "c"), ("b", "a")])
        self.assertEqual(expected, NxMixedGraph.from_str_adj(directed=directed))

    def test_is_acyclic(self):
        """Test the directed edges are acyclic."""
        example = NxMixedGraph.from_str_edges(directed=[("a", "b"), ("a", "c"), ("b", "a")])
        self.assertFalse(nx.algorithms.dag.is_directed_acyclic_graph(example.directed))

    def test_is_not_multigraph(self):
        """Test the undirected edges are not inverses of each other."""
        redundant_edges = [("a", "b"), ("b", "a")]
        directed_edges = [("a", "b")]
        expected = NxMixedGraph.from_str_edges(directed=[("a", "b")], undirected=[("a", "b")])
        actual = NxMixedGraph.from_str_edges(directed=directed_edges, undirected=redundant_edges)
        self.assertEqual(expected, actual)

    def test_subgraph(self):
        """Test generating a subgraph from a set of vertices."""
        graph = NxMixedGraph.from_str_edges(
            directed=[("X", "Y"), ("Y", "Z")],
            undirected=[("X", "Z")],
        )
        self.assertEqual(graph, graph.subgraph({X, Y, Z}))

        subgraph = NxMixedGraph.from_str_edges(directed=[("X", "Y")])
        self.assertEqual(subgraph, graph.subgraph({X, Y}))

    def test_intervention(self):
        """Test generating a subgraph based on an intervention."""
        graph = NxMixedGraph.from_str_edges(
            directed=[("X", "Y"), ("Z", "X")],
            undirected=[("X", "Z"), ("X", "Y"), ("Y", "Z")],
        )
        self.assertEqual(graph, graph.remove_in_edges(set()))

        intervened_graph = NxMixedGraph.from_str_edges(
            directed=[("X", "Y")],
            undirected=[("Z", "Y")],
        )
        self.assertEqual(intervened_graph, graph.remove_in_edges({X}))
        self.assertEqual(intervened_graph, graph.remove_in_edges(X))

        with self.assertRaises(TypeError):
            self.assertEqual(intervened_graph, graph.remove_in_edges({-X}))

    def test_remove_nodes_from(self):
        """Test generating a new graph without the given nodes."""
        graph = NxMixedGraph.from_str_edges(
            directed=[("X", "Y"), ("Z", "X")],
            undirected=[("X", "Z"), ("X", "Y"), ("Y", "Z")],
        )
        self.assertEqual(graph, graph.remove_nodes_from(set()))

        subgraph = NxMixedGraph.from_str_edges(undirected=[("Z", "Y")])
        self.assertEqual(subgraph, graph.remove_nodes_from({X}))

    def test_remove_outgoing_edges_from(self):
        """Test generating a new graph without the outgoing edgs from the given nodes."""
        graph = NxMixedGraph.from_str_edges(directed=[("X", "Y")])
        self.assertEqual(graph, graph.remove_out_edges(set()))

        graph = NxMixedGraph.from_str_edges(undirected=[("X", "Y")])
        self.assertEqual(graph, graph.remove_out_edges(set()))

        graph = NxMixedGraph.from_str_edges(directed=[("W", "X"), ("X", "Y"), ("Y", "Z")])
        expected = NxMixedGraph.from_str_edges(directed=[("W", "X"), ("Y", "Z")])
        self.assertEqual(expected, graph.remove_out_edges({X}))

    def test_ancestors_inclusive(self):
        """Test getting ancestors, inclusive."""
        graph = NxMixedGraph.from_str_edges(
            directed=[("C", "A"), ("C", "B"), ("D", "C"), ("A", "X"), ("A", "Y"), ("B", "Z")]
        )
        self.assertEqual({A, B, C, D}, graph.ancestors_inclusive({A, B}))
        self.assertEqual({A, X, Y}, graph.descendants_inclusive({A}))
        self.assertEqual({A, B, X, Y, Z}, graph.descendants_inclusive({A, B}))

        graph = NxMixedGraph.from_str_edges(
            directed=[("X", "Z"), ("Z", "Y")], undirected=[("X", "Y")]
        )
        self.assertEqual({X, Y, Z}, graph.ancestors_inclusive({Y}))
        self.assertEqual({X, Y, Z}, graph.ancestors_inclusive(Y))
        self.assertEqual({X, Z}, graph.ancestors_inclusive({Z}))
        self.assertEqual({X}, graph.ancestors_inclusive({X}))

    def test_get_c_components(self):
        """Test that get_c_components works correctly."""
        g1 = NxMixedGraph().from_str_edges(directed=[("X", "Y"), ("Z", "X"), ("Z", "Y")])
        c1 = {frozenset([X]), frozenset([Y]), frozenset([Z])}
        g2 = NxMixedGraph().from_str_edges(directed=[("X", "Y")], undirected=[("X", "Y")])
        c2 = {frozenset([X, Y])}
        g3 = NxMixedGraph().from_edges(directed=[(X, M), (M, Y)], undirected=[(X, Y)])
        c3 = {frozenset([X, Y]), frozenset([M])}
        for graph, components in [(g1, c1), (g2, c2), (g3, c3)]:
            self.assertIsInstance(graph, NxMixedGraph)
            actual_components = graph.districts()
            self.assertTrue(
                all(isinstance(component, frozenset) for component in actual_components)
            )
            self.assertTrue(
                all(
                    isinstance(node, Variable)
                    for component in actual_components
                    for node in component
                )
            )
            self.assertEqual(components, actual_components)

    def test_get_district(self):
        """Test getting districts."""
        graph = NxMixedGraph().from_edges(directed=[(X, M), (M, Y)], undirected=[(X, Y)])
        self.assertEqual(frozenset([X, Y]), graph.get_district(X))
        self.assertEqual(frozenset([X, Y]), graph.get_district(Y))
        self.assertEqual(frozenset([M]), graph.get_district(M))
        with self.assertRaises(KeyError):
            graph.get_district(Z)

    def test_counterfactual_predicate(self):
        """Test checking counterfactual graph."""
        graph = NxMixedGraph.from_edges(directed=[(X, Y)])
        self.assertFalse(graph.is_counterfactual())
        graph.raise_on_counterfactual()

        graph = NxMixedGraph.from_edges(directed=[(X @ Y, Y)])
        self.assertTrue(graph.is_counterfactual())
        with self.assertRaises(ValueError):
            graph.raise_on_counterfactual()

    def test_markov_blanket(self):
        """Test getting a markov blanket."""
        x = [Variable(f"X{i}") for i in range(12)]
        graph = NxMixedGraph.from_edges(
            directed=[
                (x[1], x[5]),
                (x[2], x[6]),
                (x[3], x[6]),
                (x[4], x[3]),
                (x[5], x[8]),
                (x[6], x[8]),
                (x[6], x[9]),
                (x[7], x[9]),
                (x[8], x[10]),
                (x[9], x[11]),
            ]
        )
        self.assertEqual(
            {x[2], x[3], x[5], x[7], x[8], x[9]},
            graph.get_markov_blanket(x[6]),
        )

    def test_disorient(self):
        """Test disorienting a graph."""
        graph = NxMixedGraph.from_edges(directed=[(X, Y)], undirected=[(Y, Z)])
        disoriented = graph.disorient()
        self.assertTrue(disoriented.has_edge(X, Y))
        self.assertTrue(disoriented.has_edge(Y, Z))

    def test_pre(self):
        """Test getting the pre-ordering for a given node or set of nodes."""
        g1 = NxMixedGraph.from_str_adj(
            directed={"1": ["2", "3"], "2": ["4", "5"], "3": ["4"], "4": ["5"]}
        )
        g1_ananke = g1.to_admg()
        g1_y0_pre = set(g1.pre(Variable("4")))
        g1_ananke_pre = set(g1_ananke.pre(vertices=["4"], top_order=g1_ananke.topological_sort()))
        g1_y0_pre = {node.name for node in g1_y0_pre}
        self.assertEqual(g1_y0_pre, g1_ananke_pre)

        g2_ananke = SARS_SMALL_GRAPH.to_admg()
        g2_y0_pre = set(SARS_SMALL_GRAPH.pre(Variable("cytok")))
        g2_ananke_pre = set(
            g2_ananke.pre(vertices=["cytok"], top_order=g2_ananke.topological_sort())
        )
        g2_y0_pre = {node.name for node in g2_y0_pre}
        self.assertEqual(g2_y0_pre, g2_ananke_pre)


class TestFixability(unittest.TestCase):
    """A test case for fixability in estimation workflows and tools."""

    def assert_mb_shielded(self, graph: NxMixedGraph):
        """Assert the graph is mb-shielded."""
        self.assertTrue(graph.to_admg().mb_shielded())
        self.assertTrue(is_markov_blanket_shielded(graph))

    def assert_mb_unshielded(self, graph: NxMixedGraph):
        """Assert the graph is not mb-shielded."""
        self.assertFalse(graph.to_admg().mb_shielded())
        self.assertFalse(is_markov_blanket_shielded(graph))

    def test_is_mb_shielded(self):
        """Test checking for markov blanket shielding."""
        # Adapted from https://gitlab.com/causal/ananke/-/blob/dev/tests/estimation/test_automated_if.py#L80-92
        graph_unshielded = NxMixedGraph.from_str_edges(
            directed=[("T", "M"), ("M", "L"), ("L", "Y")], undirected=[("M", "Y")]
        )
        with self.subTest(name="Graph 1"):
            self.assert_mb_unshielded(graph_unshielded)

        # Second test: Napkin model
        with self.subTest(name="Napkin"):
            self.assert_mb_unshielded(napkin)

        # Third test
        with self.subTest(name="Graph 3"):
            self.assert_mb_unshielded(SARS_SMALL_GRAPH)

        # Fourth test
        graph_4 = NxMixedGraph.from_str_edges(
            directed=[
                ("SOS", "Ras"),
                ("Ras", "PI3K"),
                ("Ras", "Raf"),
                ("PI3K", "AKT"),
                ("AKT", "Raf"),
            ],
            undirected=[("SOS", "PI3K")],
        )
        with self.subTest(name="Graph 4"):
            self.assert_mb_shielded(graph_4)

        # Fifth test
        graph_5 = NxMixedGraph.from_str_edges(
            directed=[
                ("Z1", "X"),
                ("X", "M1"),
                ("M1", "Y"),
                ("Z1", "Z2"),
                ("Z2", "Z3"),
                ("Z3", "Y"),
                ("Z2", "M1"),
            ],
            undirected=[("Z1", "X"), ("Z2", "M1")],
        )
        with self.subTest(name="Graph 5"):
            self.assert_mb_unshielded(graph_5)

        # Sixth test
        graph_6 = NxMixedGraph.from_str_edges(
            directed=[("Z1", "X"), ("X", "M1"), ("M1", "Y"), ("Z1", "Y")]
        )
        with self.subTest(name="Graph 6"):
            self.assert_mb_shielded(graph_6)

        # Seventh test
        graph_7 = NxMixedGraph.from_str_edges(directed=[("A", "B"), ("B", "C")])
        with self.subTest(name="Graph 6"):
            self.assert_mb_shielded(graph_7)

    def assert_a_fixable(self, graph: NxMixedGraph, treatment: Variable):
        """Assert that the graph is a-fixable."""
        self.assertTrue(_ananke_a_fixable(graph, treatment))
        self.assertTrue(is_a_fixable(graph, treatment))

    def assert_not_a_fixable(self, graph: NxMixedGraph, treatment: Variable):
        """Assert that the graph is not a-fixable."""
        self.assertFalse(_ananke_a_fixable(graph, treatment))
        self.assertFalse(is_a_fixable(graph, treatment))

    def test_is_a_fixable(self):
        """Test checking for a-fixability.

        Graphs 9, 10, and 11 are from https://gitlab.com/causal/ananke/-/blob/dev/tests/\
        estimation/test_counterfactual_mean.py?ref_type=heads#L20-47
        """
        graph_1 = NxMixedGraph.from_str_edges(
            directed=[("T", "M"), ("M", "L"), ("L", "Y")], undirected=[("M", "Y")]
        )
        treatment_1 = Variable("T")
        self.assert_a_fixable(graph_1, treatment_1)

        self.assert_not_a_fixable(napkin, X)

        graph_3 = NxMixedGraph.from_str_edges(
            directed=[
                ("ADAM17", "EGFR"),
                ("ADAM17", "TNF"),
                ("ADAM17", "Sil6r"),
                ("EGFR", "cytok"),
                ("TNF", "cytok"),
                ("Sil6r", "IL6STAT3"),
                ("IL6STAT3", "cytok"),
            ],
            undirected=[
                ("ADAM17", "cytok"),
                ("ADAM17", "Sil6r"),
                ("EGFR", "TNF"),
                ("EGFR", "IL6STAT3"),
            ],
        )
        treatment_3 = Variable("EGFR")
        self.assert_a_fixable(graph_3, treatment_3)

        graph_4 = NxMixedGraph.from_str_edges(
            directed=[
                ("SOS", "Ras"),
                ("Ras", "PI3K"),
                ("Ras", "Raf"),
                ("PI3K", "AKT"),
                ("AKT", "Raf"),
            ],
            undirected=[("SOS", "PI3K")],
        )
        treatment_4 = Variable("PI3K")
        self.assert_a_fixable(graph_4, treatment_4)

        graph_5 = NxMixedGraph.from_str_edges(
            directed=[
                ("Z1", "X"),
                ("X", "M1"),
                ("M1", "Y"),
                ("Z1", "Z2"),
                ("Z2", "Z3"),
                ("Z3", "Y"),
                ("Z2", "M1"),
            ],
            undirected=[("Z1", "X"), ("Z2", "M1")],
        )
        self.assert_a_fixable(graph_5, X)

        graph_6 = NxMixedGraph.from_str_edges(
            directed=[("Z1", "X"), ("X", "M1"), ("M1", "Y"), ("Z1", "Y")], undirected=[]
        )
        self.assert_a_fixable(graph_6, X)

        graph_7 = NxMixedGraph.from_str_edges(directed=[("A", "B"), ("B", "C")], undirected=[])
        treatment_7 = Variable("A")
        self.assert_a_fixable(graph_7, treatment_7)

        graph_8 = graph_4
        treatment_8 = Variable("SOS")
        self.assert_not_a_fixable(graph_8, treatment_8)

    def assert_p_fixable(self, graph: NxMixedGraph, treatment: Variable):
        """Assert that the graph is p-fixable."""
        self.assertTrue(_ananke_p_fixable(graph, treatment))
        self.assertTrue(is_p_fixable(graph, treatment))

    def assert_not_p_fixable(self, graph: NxMixedGraph, treatment: Variable):
        """Assert that the graph is not p-fixable."""
        self.assertFalse(_ananke_p_fixable(graph, treatment))
        self.assertFalse(is_p_fixable(graph, treatment))

    def test_is_p_fixable(self):
        """Test checking for p-fixability.

        .. seealso:: https://gitlab.com/causal/ananke/-/blob/dev/tests/\
        estimation/test_counterfactual_mean.py?ref_type=heads#L151-212
        """
        graph_1 = NxMixedGraph.from_str_edges(
            directed=[("T", "M"), ("M", "L"), ("L", "Y")], undirected=[("M", "Y")]
        )
        treatment_1 = Variable("T")
        with self.subTest(name="Graph 1"):
            self.assert_p_fixable(graph_1, treatment_1)

        with self.subTest(name="Napkin"):
            self.assert_not_p_fixable(napkin, X)

        treatment_3 = Variable("EGFR")
        with self.subTest(name="Graph 3"):
            self.assert_p_fixable(SARS_SMALL_GRAPH, treatment_3)

        graph_4 = NxMixedGraph.from_str_edges(
            directed=[
                ("SOS", "Ras"),
                ("Ras", "PI3K"),
                ("Ras", "Raf"),
                ("PI3K", "AKT"),
                ("AKT", "Raf"),
            ],
            undirected=[("SOS", "PI3K")],
        )
        treatment_4 = Variable("PI3K")
        with self.subTest(name="Graph 4"):
            self.assert_p_fixable(graph_4, treatment_4)

        graph_5 = NxMixedGraph.from_str_edges(
            directed=[
                ("Z1", "X"),
                ("X", "M1"),
                ("M1", "Y"),
                ("Z1", "Z2"),
                ("Z2", "Z3"),
                ("Z3", "Y"),
                ("Z2", "M1"),
            ],
            undirected=[("Z1", "X"), ("Z2", "M1")],
        )
        with self.subTest(name="Graph 5"):
            self.assert_p_fixable(graph_5, X)

        graph_6 = NxMixedGraph.from_str_edges(
            directed=[("Z1", "X"), ("X", "M1"), ("M1", "Y"), ("Z1", "Y")], undirected=[]
        )
        with self.subTest(name="Graph 6"):
            self.assert_p_fixable(graph_6, X)

        graph_7 = NxMixedGraph.from_str_edges(directed=[("A", "B"), ("B", "C")], undirected=[])
        treatment_7 = Variable("A")
        with self.subTest(name="Graph 7"):
            self.assert_p_fixable(graph_7, treatment_7)

        graph_8 = graph_4
        treatment_8 = Variable("SOS")
        with self.subTest(name="Graph 8"):
            self.assert_p_fixable(graph_8, treatment_8)

        graph_9 = NxMixedGraph.from_str_edges(
            directed=[
                ("C", "T"),
                ("C", "M"),
                ("C", "L"),
                ("C", "Y"),
                ("T", "M"),
                ("M", "L"),
                ("M", "Y"),
                ("L", "Y"),
            ],
            undirected=[("T", "L"), ("T", "Y")],
        )
        treatment_9 = Variable("T")
        with self.subTest(name="Graph 9"):
            self.assert_p_fixable(graph_9, treatment_9)

        graph_10 = NxMixedGraph.from_str_edges(
            directed=[
                ("C", "T"),
                ("C", "M"),
                ("C", "L"),
                ("C", "Y"),
                ("T", "M"),
                ("M", "L"),
                ("T", "Y"),
                ("L", "Y"),
            ],
            undirected=[("T", "L"), ("M", "Y")],
        )
        treatment_10 = Variable("T")
        with self.subTest(name="Graph 10"):
            self.assert_p_fixable(graph_10, treatment_10)

        graph_11 = NxMixedGraph.from_str_edges(
            directed=[
                ("C1", "T"),
                ("C1", "L"),
                ("C2", "M"),
                ("C2", "L"),
                ("C2", "Y"),
                ("T", "M"),
                ("M", "L"),
                ("L", "Y"),
            ],
            undirected=[("Z1", "C1"), ("Z2", "C2"), ("T", "L")],
        )
        treatment_11 = Variable("T")
        with self.subTest(name="Graph 11"):
            self.assert_p_fixable(graph_11, treatment_11)

        graph_12 = NxMixedGraph.from_str_edges(
            directed=[
                ("C2", "T"),
                ("C2", "M1"),
                ("C2", "M2"),
                ("C2", "Y"),
                ("C1", "T"),
                ("C1", "M2"),
                ("T", "M1"),
                ("M1", "M2"),
                ("M2", "Y"),
                ("T", "Y"),
            ],
            undirected=[("T", "M2"), ("M1", "Y")],
        )
        treatment_12 = Variable("T")
        with self.subTest(name="Graph 12"):
            self.assert_p_fixable(graph_12, treatment_12)

        graph_13 = NxMixedGraph.from_str_edges(
            directed=[
                ("ViralLoad", "Income"),
                ("ViralLoad", "T"),
                ("ViralLoad", "Toxicity"),
                ("Education", "Income"),
                ("Education", "T"),
                ("Education", "Toxicity"),
                ("Income", "Insurance"),
                ("Insurance", "T"),
                ("T", "Toxicity"),
                ("Toxicity", "CD4"),
                ("T", "CD4"),
            ],
            undirected=[
                ("Income", "T"),
                ("Insurance", "ViralLoad"),
                ("Education", "CD4"),
                ("Income", "CD4"),
            ],
        )
        treatment_13 = Variable("T")
        with self.subTest(name="Graph 13"):
            self.assert_not_p_fixable(graph_13, treatment_13)


def _ananke_a_fixable(graph: NxMixedGraph, treatment: Variable) -> bool:
    admg = graph.to_admg()
    return 1 == len(admg.district(treatment.name).intersection(admg.descendants([treatment.name])))


def _ananke_p_fixable(graph: NxMixedGraph, treatment: Variable) -> bool:
    admg = graph.to_admg()
    return 0 == len(admg.district(treatment.name).intersection(admg.children([treatment.name])))
